#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import copy
import pprint

from collections.abc import Mapping, Sequence
from typing import Dict

from twisted.internet.defer import Deferred, inlineCallbacks

from autobahn import util
from autobahn.wamp.types import SessionIdent
from autobahn.util import hl, hlid, hltype, hluserid, hlval

from crossbar.common.checkconfig import check_dict_args, check_realm_name, check_connecting_transport
from crossbar.common.twisted.endpoint import create_connecting_endpoint_from_config

from autobahn.wamp.types import SubscribeOptions, PublishOptions, RegisterOptions, CallOptions, ComponentConfig
from autobahn.wamp.message import Event, Invocation
from autobahn.wamp.exception import ApplicationError, TransportLost
from autobahn.twisted.wamp import ApplicationSession, ApplicationRunner

from txaio import make_logger, time_ns

__all__ = (
    'RLink',
    'RLinkConfig',
    'RLinkManager',
)


class BridgeSession(ApplicationSession):

    log = make_logger()

    def __init__(self, config):
        ApplicationSession.__init__(self, config)

        self._subs = {}
        # registration-id's of remote registrations from an rlink
        self._regs = {}

        self._exclude_authid = None
        self._exclude_authrole = None

    def onMessage(self, msg):
        if msg._router_internal is not None:
            if isinstance(msg, Event):
                msg.publisher, msg.publisher_authid, msg.publisher_authrole = msg._router_internal
            elif isinstance(msg, Invocation):
                msg.caller, msg.caller_authid, msg.caller_authrole = msg._router_internal
        return super(BridgeSession, self).onMessage(msg)

    @inlineCallbacks
    def _setup_event_forwarding(self, other):

        self.log.debug(
            "setup event forwarding between {me} and {other} (exclude_authid={exclude_authid}, exclude_authrole={exclude_authrole})",
            exclude_authid=self._exclude_authid,
            exclude_authrole=self._exclude_authrole,
            me=self._session_id,
            other=other)

        @inlineCallbacks
        def on_subscription_create(sub_id, sub_details, details=None):
            """
            Event handler fired when a new subscription was created on this router.

            The handler will then also subscribe on the other router, and when receiving
            events, re-publish those on this router.

            :param sub_id:
            :param sub_details:
            :param details:
            :return:
            """
            if sub_id in self._subs:
                # this should not happen actually, but not sure ..
                self.log.error('on_subscription_create: sub ID {sub_id} already in map {method}',
                               sub_id=sub_id,
                               method=hltype(BridgeSession._setup_event_forwarding))
                return

            self._subs[sub_id] = sub_details

            uri = sub_details['uri']
            ERR_MSG = [None]

            @inlineCallbacks
            def on_event(*args, **kwargs):
                assert 'details' in kwargs
                details = kwargs.pop('details')
                options = kwargs.pop('options', None)

                self.log.debug(
                    'Received event on uri={uri}, options={options} (publisher={publisher}, publisher_authid={publisher_authid}, publisher_authrole={publisher_authrole}, forward_for={forward_for})',
                    uri=uri,
                    options=options,
                    publisher=details.publisher,
                    publisher_authid=details.publisher_authid,
                    publisher_authrole=details.publisher_authrole,
                    forward_for=details.forward_for)

                assert details.publisher is not None
                this_forward = {
                    'session': details.publisher,
                    'authid': details.publisher_authid,
                    'authrole': details.publisher_authrole,
                }

                if details.forward_for:
                    # the event comes already forwarded from a router node
                    if len(details.forward_for) >= 0:
                        self.log.debug('SKIP! already forwarded')
                        return

                    forward_for = copy.deepcopy(details.forward_for)
                    forward_for.append(this_forward)
                else:
                    forward_for = [this_forward]

                options = PublishOptions(acknowledge=True,
                                         exclude_me=True,
                                         exclude_authid=self._exclude_authid,
                                         exclude_authrole=self._exclude_authrole,
                                         forward_for=forward_for)

                try:
                    yield self.publish(uri, *args, options=options, **kwargs)
                except TransportLost:
                    return
                except ApplicationError as e:
                    if e.error not in ['wamp.close.normal']:
                        self.log.warn('FAILED TO PUBLISH 1: {} {}'.format(type(e), str(e)))
                    return
                except Exception as e:
                    if not ERR_MSG[0]:
                        self.log.warn('FAILED TO PUBLISH 2: {} {}'.format(type(e), str(e)))
                        ERR_MSG[0] = True
                    return

                self.log.debug(
                    "RLink forward-published event {dir} (options={options})",
                    dir=self.DIR,
                    options=options,
                )

            sub = yield other.subscribe(on_event, uri, options=SubscribeOptions(details=True))
            self._subs[sub_id]['sub'] = sub

            self.log.debug(
                "created forwarding subscription: me={me} other={other} sub_id={sub_id} sub_details={sub_details} details={details}",
                me=self._session_id,
                other=other,
                sub_id=sub_id,
                sub_details=sub_details,
                details=details,
            )

        # listen to when a subscription is removed from the router
        #
        @inlineCallbacks
        def on_subscription_delete(session_id, sub_id, details=None):
            self.log.debug(
                "Subscription deleted: {me} {session} {sub_id} {details}",
                me=self,
                session=session_id,
                sub_id=sub_id,
                details=details,
            )

            sub_details = self._subs.get(sub_id, None)
            if not sub_details:
                self.log.debug("subscription not tracked - huh??")
                return

            uri = sub_details['uri']

            yield self._subs[sub_id]['sub'].unsubscribe()

            del self._subs[sub_id]

            self.log.debug("{other} unsubscribed from {uri}".format(other=other, uri=uri))

        @inlineCallbacks
        def forward_current_subs():
            # get current subscriptions on the router
            #
            subs = yield self.call("wamp.subscription.list")
            for sub_id in subs['exact']:
                sub = yield self.call("wamp.subscription.get", sub_id)

                if not sub['uri'].startswith("wamp."):
                    yield on_subscription_create(sub_id, sub)

        @inlineCallbacks
        def on_remote_join(_session, _details):
            yield forward_current_subs()

        if self.IS_REMOTE_LEG:
            yield forward_current_subs()
        else:
            # from the local leg, don't try to forward events on the
            # remote leg unless the remote session is established.
            other.on('join', on_remote_join)

        # listen to when new subscriptions are created on the local router
        yield self.subscribe(on_subscription_create,
                             "wamp.subscription.on_create",
                             options=SubscribeOptions(details_arg="details"))

        yield self.subscribe(on_subscription_delete,
                             "wamp.subscription.on_delete",
                             options=SubscribeOptions(details_arg="details"))

        self.log.debug("{me}: event forwarding setup done", me=self)

    @inlineCallbacks
    def _setup_invocation_forwarding(self, other: ApplicationSession):

        self.log.info(
            "setup invocation forwarding between {me} and {other} (exclude_authid={exclude_authid}, exclude_authrole={exclude_authrole})",
            exclude_authid=self._exclude_authid,
            exclude_authrole=self._exclude_authrole,
            me=self,
            other=other)

        # called when a registration is created on the local router
        @inlineCallbacks
        def on_registration_create(reg_session, reg_details, details=None):
            """
            Event handler fired when a new registration was created on this router.

            The handler will then also register on the other router, and when receiving
            calls, re-issue those on this router.

            :param reg_session:
            :param reg_details:
            :param details:
            :return:
            """
            if reg_details['uri'].startswith("wamp."):
                return

            if reg_details['id'] in self._regs:
                # this should not happen actually, but not sure ..
                self.log.error('on_registration_create: reg ID {reg_id} already in map {method}',
                               reg_id=reg_details['id'],
                               method=hltype(BridgeSession._setup_invocation_forwarding))
                return

            self._regs[reg_details['id']] = reg_details
            self._regs[reg_details['id']]['reg'] = None

            uri = reg_details['uri']
            ERR_MSG = [None]

            @inlineCallbacks
            def on_call(*args, **kwargs):

                assert 'details' in kwargs

                details = kwargs.pop('details')
                options = kwargs.pop('options', None)

                if details.caller is None or details.caller_authrole is None or details.caller_authid is None:
                    raise RuntimeError("Internal error attempting rlink forwarding")

                self.log.info(
                    'Received invocation on uri={uri}, options={options} (caller={caller}, caller_authid={caller_authid}, caller_authrole={caller_authrole}, forward_for={forward_for})',
                    uri=uri,
                    options=options,
                    caller=details.caller,
                    caller_authid=details.caller_authid,
                    caller_authrole=details.caller_authrole,
                    forward_for=details.forward_for)

                this_forward = {
                    'session': details.caller,
                    'authid': details.caller_authrole,
                    'authrole': details.caller_authrole,
                }

                if details.forward_for:
                    # the call comes already forwarded from a router node ..
                    if len(details.forward_for) >= 0:
                        self.log.debug('SKIP! already forwarded')
                        return

                    forward_for = copy.deepcopy(details.forward_for)
                    forward_for.append(this_forward)
                else:
                    forward_for = [this_forward]

                options = CallOptions(forward_for=forward_for)

                try:
                    result = yield self.call(uri, *args, options=options, **kwargs)
                except TransportLost:
                    return
                except ApplicationError as e:
                    if e.error not in ['wamp.close.normal']:
                        self.log.warn('FAILED TO CALL 1: {} {}'.format(type(e), str(e)))
                    return
                except Exception as e:
                    if not ERR_MSG[0]:
                        self.log.warn('FAILED TO CALL 2: {} {}'.format(type(e), str(e)))
                        ERR_MSG[0] = True
                    return

                self.log.info(
                    "RLink forward-invoked call {dir} (options={options})",
                    dir=self.DIR,
                    options=options,
                )
                return result

            try:
                reg = yield other.register(on_call,
                                           uri,
                                           options=RegisterOptions(
                                               details_arg='details',
                                               invoke=reg_details.get('invoke', None),
                                           ))
            except Exception as e:
                # FIXME: partially fixes https://github.com/crossbario/crossbar/issues/1894,
                #  however we need to make sure this situation never happens.
                if isinstance(e, ApplicationError) and e.error == 'wamp.error.procedure_already_exists':
                    other_leg = 'local' if self.IS_REMOTE_LEG else 'remote'
                    self.log.debug(f"on_registration_create: tried to register procedure {uri} on {other_leg} "
                                   f"session but it's already registered.")
                    return
                raise Exception("fatal: could not forward-register '{}'".format(uri))

            # so ... if, during that "yield" above while we register
            # on the "other" router, *this* router may have already
            # un-registered. If that happened, our registration will
            # be gone, so we immediately un-register on the other side
            if reg_details['id'] not in self._regs:
                self.log.info("registration already gone: {uri}", uri=reg_details['uri'])
                yield reg.unregister()
            else:
                self._regs[reg_details['id']]['reg'] = reg

            self.log.info(
                "created forwarding registration: me={me} other={other} reg_id={reg_id} reg_details={reg_details} details={details} reg_session={reg_session}",
                me=self._session_id,
                other=other._session_id,
                reg_id=reg_details['id'],
                reg_details=reg_details,
                details=details,
                reg_session=reg_session,
            )

        # called when a registration is removed from the local router
        @inlineCallbacks
        def on_registration_delete(session_id, reg_id, details=None):
            self.log.info(
                "Registration deleted: {me} {session} {reg_id} {details}",
                me=self,
                session=session_id,
                reg_id=reg_id,
                details=details,
            )

            reg_details = self._regs.get(reg_id, None)
            if not reg_details:
                self.log.info("registration not tracked - huh??")
                return

            uri = reg_details['uri']

            reg = self._regs[reg_id]['reg']
            if reg is None:
                # see above; we might have un-registered here before
                # we got an answer from the other router
                self.log.info("registration has no 'reg'")
            else:
                yield reg.unregister()

            del self._regs[reg_id]

            self.log.info("{other} unsubscribed from {uri}".format(other=other, uri=uri))

        @inlineCallbacks
        def register_current():
            # get current registrations on the router
            regs = yield self.call("wamp.registration.list")
            for reg_id in regs['exact']:
                reg = yield self.call("wamp.registration.get", reg_id)
                assert reg['id'] == reg_id, "Logic error, registration IDs don't match"
                yield on_registration_create(self._session_id, reg)

        @inlineCallbacks
        def on_remote_join(_session, _details):
            yield register_current()

        def on_remote_leave(_session, _details):
            # The remote session has ended, clear registration records.
            # Clearing this dictionary helps avoid the case where
            # local procedures are not registered on the remote leg
            # on reestablishment of remote session.
            # See: https://github.com/crossbario/crossbar/issues/1909
            self._regs = {}

        if self.IS_REMOTE_LEG:
            yield register_current()
        else:
            # from the local leg, don't try to register procedures on the
            # remote leg unless the remote session is established.
            # This avoids issues where in-router components register procedures
            # on startup and when the rlink is setup, the local leg tries to
            # register procedures on the remote leg, even though the connection
            # hasn't established.
            # See: https://github.com/crossbario/crossbar/issues/1895
            other.on('join', on_remote_join)
            other.on('leave', on_remote_leave)

        # listen to when new registrations are created on the local router
        yield self.subscribe(on_registration_create,
                             "wamp.registration.on_create",
                             options=SubscribeOptions(details_arg="details"))

        # listen to when a registration is removed from the local router
        yield self.subscribe(on_registration_delete,
                             "wamp.registration.on_delete",
                             options=SubscribeOptions(details_arg="details"))

        self.log.info("{me}: call forwarding setup done", me=self._session_id)


class RLinkLocalSession(BridgeSession):
    """
    This session is the local leg of the router-to-router link and runs embedded inside the local router.
    """

    log = make_logger()

    IS_REMOTE_LEG = False

    # direction in which events are flowing (published) via this session
    DIR = hl('from remote to local', color='yellow', bold=True)

    def onConnect(self):
        self.log.info('{klass}.onConnect()', klass=self.__class__.__name__)
        # _BridgeSession.onConnect(self)
        authextra = {'rlink': self.config.extra['rlink']}
        self.join(self.config.realm, authid=self.config.extra['rlink'], authextra=authextra)
        self._tracker = self.config.extra['tracker']

    @inlineCallbacks
    def onJoin(self, details):
        assert self.config.extra and 'on_ready' in self.config.extra
        assert self.config.extra and 'other' in self.config.extra

        remote = self.config.extra['other']
        assert isinstance(remote, RLinkRemoteSession)

        self._exclude_authid = self.config.extra.get('exclude_authid', None)
        self._exclude_authrole = self.config.extra.get('exclude_authrole', None)

        # setup local->remote event forwarding
        forward_events = self.config.extra.get('forward_events', False)
        if forward_events:
            yield self._setup_event_forwarding(remote)

        # setup local->remote invocation forwarding
        forward_invocations = self.config.extra.get('forward_invocations', False)
        if forward_invocations:
            yield self._setup_invocation_forwarding(remote)

        self.log.debug(
            'Router link local session ready (forward_events={forward_events}, forward_invocations={forward_invocations}, realm={realm}, authid={authid}, authrole={authrole}, session={session}) {method}',
            method=hltype(RLinkLocalSession.onJoin),
            forward_events=hluserid(forward_events),
            forward_invocations=hluserid(forward_invocations),
            realm=hluserid(details.realm),
            authid=hluserid(details.authid),
            authrole=hluserid(details.authrole),
            session=hlid(details.session))

        on_ready = self.config.extra.get('on_ready', None)
        if on_ready and not on_ready.called:
            self.config.extra['on_ready'].callback(self)

    def onLeave(self, details):
        self.log.warn(
            'Router link local session down! (realm={realm}, authid={authid}, authrole={authrole}, session={session}, details={details}) {method}',
            method=hltype(RLinkLocalSession.onLeave),
            realm=hluserid(self.config.realm),
            authid=hluserid(self._authid),
            authrole=hluserid(self._authrole),
            details=details,
            session=hlid(self._session_id))

        BridgeSession.onLeave(self, details)


class RLinkRemoteSession(BridgeSession):
    """
    This session is the remote leg of the router-to-router link.
    """

    log = make_logger()

    IS_REMOTE_LEG = True

    # directory in which events are flowing (published via this session
    DIR = hl('from local to remote', color='yellow', bold=True)

    def __init__(self, config):
        BridgeSession.__init__(self, config)

        # import here to resolve import dependency issues
        from crossbar.worker.router import RouterController

        self._subs = {}
        self._rlink_manager: RLinkManager = self.config.extra['rlink_manager']
        self._router_controller: RouterController = self._rlink_manager.controller

    # FIXME: async? see below
    def onConnect(self):
        self.log.info('{func}() ...', func=hltype(self.onConnect))

        authid = self.config.extra.get('authid', None)
        authrole = self.config.extra.get('authrole', None)
        authextra = self.config.extra.get('authextra', {})

        # FIXME: use cryptosign-proxy
        authmethods = ['cryptosign']

        # use WorkerController.get_public_key to call node controller
        # FIXME: the following does _not_ work with onConnect (?!)
        # _public_key = await self._router_controller.get_public_key()

        def actually_join(_public_key):
            authextra.update({
                # forward the client pubkey: this allows us to omit authid as
                # the router can identify us with the pubkey already
                'pubkey': _public_key,

                # not yet implemented. a public key the router should provide
                # a trustchain for its public key. the trustroot can eg be
                # hard-coded in the client, or come from a command line option.
                'trustroot': None,

                # not yet implemented. for authenticating the router, this
                # challenge will need to be signed by the router and send back
                # in AUTHENTICATE for client to verify. A string with a hex
                # encoded 32 bytes random value.
                'challenge': None,

                # https://tools.ietf.org/html/rfc5929
                'channel_binding': 'tls-unique'
            })

            self.log.info(
                '{func} joining with realm="{realm}", authmethods={authmethods}, authid="{authid}", authrole="{authrole}", authextra={authextra}',
                func=hltype(self.onConnect),
                realm=hlval(self.config.realm),
                authmethods=hlval(authmethods),
                authid=hlval(authid),
                authrole=hlval(authrole),
                authextra=authextra)

            self.join(self.config.realm,
                      authmethods=authmethods,
                      authid=authid,
                      authrole=authrole,
                      authextra=authextra)

        res = self._rlink_manager._controller.get_public_key()
        res.addCallback(actually_join)
        self.log.info('{func}() done (res={res}).', func=hltype(self.onConnect), res=res)
        return res

    # FIXME: async? see below
    def onChallenge(self, challenge):
        self.log.debug('{func}(challenge={challenge})', func=hltype(self.onChallenge), challenge=challenge)

        if challenge.method == 'cryptosign':
            # alright, we've got a challenge from the router.

            # sign the challenge with our private key.
            channel_id_type = 'tls-unique'
            channel_id_map = self._router_controller._transport.transport_details.channel_id
            if channel_id_type in channel_id_map:
                channel_id = channel_id_map[channel_id_type]
            else:
                channel_id = None
                channel_id_type = None

            # use WorkerController.get_public_key to call node controller
            # FIXME: await?
            signed_challenge = self._router_controller.sign_challenge(challenge, channel_id, channel_id_type)

            # send back the signed challenge for verification
            return signed_challenge

        else:
            raise Exception(
                'internal error: we asked to authenticate using wamp-cryptosign, but now received a challenge for {}'.
                format(challenge.method))

    @inlineCallbacks
    def onJoin(self, details):
        self.log.debug('{klass}.onJoin(details={details})', klass=self.__class__.__name__, details=details)

        assert self.config.extra and 'on_ready' in self.config.extra
        assert self.config.extra and 'other' in self.config.extra

        local = self.config.extra['other']
        assert isinstance(local, RLinkLocalSession)
        local._tracker.connected = True

        self._exclude_authid = self.config.extra.get('exclude_authid', None)
        self._exclude_authrole = self.config.extra.get('exclude_authrole', None)

        # setup remote->local event forwarding
        forward_events = self.config.extra.get('forward_events', False)
        if forward_events:
            yield self._setup_event_forwarding(local)

        # setup remote->local invocation forwarding
        forward_invocations = self.config.extra.get('forward_invocations', False)
        if forward_invocations:
            yield self._setup_invocation_forwarding(local)

        self.log.info(
            '{klass}.onJoin(): rlink remote session ready (forward_events={forward_events}, forward_invocations={forward_invocations}, realm={realm}, authid={authid}, authrole={authrole}, session={session}) {method}',
            klass=self.__class__.__name__,
            method=hltype(RLinkRemoteSession.onJoin),
            forward_events=hluserid(forward_events),
            forward_invocations=hluserid(forward_invocations),
            realm=hluserid(details.realm),
            authid=hluserid(details.authid),
            authrole=hluserid(details.authrole),
            session=hlid(details.session))

        # we are ready!
        on_ready = self.config.extra.get('on_ready', None)
        if on_ready and not on_ready.called:
            self.config.extra['on_ready'].callback(self)

    @inlineCallbacks
    def onLeave(self, details):
        # When the rlink is going down, make sure to unsubscribe to
        # all events that are subscribed on the local-leg.
        # This avoids duplicate events that would otherwise arrive
        # See: https://github.com/crossbario/crossbar/issues/1916
        for k, v in self._subs.items():
            if v['sub'].active:
                yield v['sub'].unsubscribe()

        self._subs = {}

        self.config.extra['other']._tracker.connected = False
        self.log.warn(
            '{klass}.onLeave(): rlink remote session left! (realm={realm}, authid={authid}, authrole={authrole}, session={session}, details={details}) {method}',
            klass=self.__class__.__name__,
            method=hltype(RLinkLocalSession.onLeave),
            realm=hluserid(self.config.realm),
            authid=hluserid(self._authid),
            authrole=hluserid(self._authrole),
            session=hlid(self._session_id),
            details=details)

        BridgeSession.onLeave(self, details)


class RLink(object):
    def __init__(self, id, config, started=None, started_by=None, local=None, remote=None):
        assert type(id) == str
        assert isinstance(config, RLinkConfig)
        assert started is None or type(started) == int
        assert started_by is None or isinstance(started_by, RLinkConfig)
        assert local is None or isinstance(local, RLinkLocalSession)
        assert remote is None or isinstance(remote, RLinkLocalSession)

        # link ID
        self.id = id

        # link config: RLinkConfig
        self.config = config

        # when was it started: epoch time in ns
        self.started = started

        # who started this link: SessionIdent
        self.started_by = started_by

        # local session: RLinkLocalSession
        self.local = local

        # remote session: RLinkRemoteSession
        self.remote = remote

        # updated by the session
        self.connected = False

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        obj = {
            'id': self.id,
            'config': self.config.marshal() if self.config else None,
            'started': self.started,
            'started_by': self.started_by.marshal() if self.started_by else None,
            'connected': self.connected,
        }
        return obj


class RLinkConfig(object):
    def __init__(self, realm, transport, authid, exclude_authid, forward_local_events, forward_remote_events,
                 forward_local_invocations, forward_remote_invocations):
        """

        :param realm: The remote router realm.
        :type realm: str

        :param transport: The transport for connecting to the remote router.
        :type transport:
        """
        self.realm = realm
        self.transport = transport
        self.authid = authid
        self.exclude_authid = exclude_authid
        self.forward_local_events = forward_local_events
        self.forward_remote_events = forward_remote_events
        self.forward_local_invocations = forward_local_invocations
        self.forward_remote_invocations = forward_remote_invocations

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        obj = {
            'realm': self.realm,
            'transport': self.transport,
            'authid': self.authid,
            'exclude_authid': self.exclude_authid,
            'forward_local_events': self.forward_local_events,
            'forward_remote_events': self.forward_remote_events,
            'forward_local_invocations': self.forward_local_invocations,
            'forward_remote_invocations': self.forward_remote_invocations,
        }
        return obj

    @staticmethod
    def parse(personality, obj, id=None):
        """
        Parses a generic object (eg a dict) into a typed
        object of this class.

        :param obj: The generic object to parse.
        :type obj: dict

        :returns: Router link configuration
        :rtype: :class:`crossbar.edge.worker.rlink.RLinkConfig`
        """
        # assert isinstance(personality, Personality)
        assert type(obj) == dict
        assert id is None or type(id) == str

        if id:
            obj['id'] = id

        check_dict_args(
            {
                'id': (False, [str]),
                'realm': (True, [str]),
                'transport': (True, [Mapping]),
                'authid': (False, [str]),
                'exclude_authid': (False, [Sequence]),
                'forward_local_events': (False, [bool]),
                'forward_remote_events': (False, [bool]),
                'forward_local_invocations': (False, [bool]),
                'forward_remote_invocations': (False, [bool]),
            }, obj, 'router link configuration')

        realm = obj['realm']
        authid = obj.get('authid', None)
        exclude_authid = obj.get('exclude_authid', [])
        for aid in exclude_authid:
            assert type(aid) == str
        forward_local_events = obj.get('forward_local_events', True)
        forward_remote_events = obj.get('forward_remote_events', True)
        forward_local_invocations = obj.get('forward_local_invocations', True)
        forward_remote_invocations = obj.get('forward_remote_invocations', True)
        transport = obj['transport']

        check_realm_name(realm)
        check_connecting_transport(personality, transport)

        config = RLinkConfig(
            realm=realm,
            transport=transport,
            authid=authid,
            exclude_authid=exclude_authid,
            forward_local_events=forward_local_events,
            forward_remote_events=forward_remote_events,
            forward_local_invocations=forward_local_invocations,
            forward_remote_invocations=forward_remote_invocations,
        )

        return config


class RLinkManager(object):
    """
    Router-to-router links manager.
    """
    log = make_logger()

    def __init__(self, realm, controller):
        """

        :param realm: The (local) router realm this object is managing links for.
        :param controller: The router controller this rlink is running under.
        """
        # import here to resolve import dependency issues
        from crossbar.edge.worker.router import ExtRouterRealm
        from crossbar.worker.router import RouterController

        self._realm: ExtRouterRealm = realm
        self._controller: RouterController = controller

        # map: link_id -> RLink
        self._links: Dict[str, RLink] = {}

    @property
    def realm(self):
        return self._realm

    @property
    def controller(self):
        return self._controller

    def __getitem__(self, link_id):
        if link_id in self._links:
            return self._links[link_id]
        else:
            raise KeyError('no router link with ID "{}"'.format(link_id))

    def __contains__(self, link_id):
        return link_id in self._links

    def __len__(self):
        return len(self._links)

    def __setitem__(self, item, value):
        raise Exception('__setitem__ not supported on this class')

    def __delitem__(self, item):
        raise Exception('__delitem__ not supported on this class')

    def keys(self):
        return self._links.keys()

    @inlineCallbacks
    def start_link(self, link_id, link_config, caller):
        assert type(link_id) == str
        assert isinstance(link_config, RLinkConfig)
        assert isinstance(caller, SessionIdent)

        if link_id in self._links:
            raise ApplicationError('crossbar.error.already_running', 'router link {} already running'.format(link_id))

        # setup local session
        #
        local_extra = {
            'other': None,
            'on_ready': Deferred(),
            'rlink': link_id,
            'forward_events': link_config.forward_local_events,
            'forward_invocations': link_config.forward_local_invocations,
        }
        local_realm = self._realm.config['name']

        local_authid = link_config.authid or util.generate_serial_number()
        local_authrole = 'trusted'
        local_config = ComponentConfig(local_realm, local_extra)
        local_session = RLinkLocalSession(local_config)

        # setup remote session
        #
        remote_extra = {
            'rlink_manager': self,
            'other': None,
            'on_ready': Deferred(),
            'authid': link_config.authid,
            'exclude_authid': link_config.exclude_authid,
            'forward_events': link_config.forward_remote_events,
            'forward_invocations': link_config.forward_remote_invocations,
        }
        remote_realm = link_config.realm
        remote_config = ComponentConfig(remote_realm, remote_extra)
        remote_session = RLinkRemoteSession(remote_config)

        # cross-connect the two sessions
        #
        local_extra['other'] = remote_session
        remote_extra['other'] = local_session

        # the rlink
        #
        rlink = RLink(link_id, link_config)
        self._links[link_id] = rlink
        local_extra['tracker'] = rlink

        # create connecting client endpoint
        #
        connecting_endpoint = create_connecting_endpoint_from_config(link_config.transport['endpoint'],
                                                                     self._controller.cbdir, self._controller._reactor,
                                                                     self.log)
        try:
            # connect the local session
            #
            self._realm.controller._router_session_factory.add(local_session,
                                                               self._realm.router,
                                                               authid=local_authid,
                                                               authrole=local_authrole,
                                                               authextra=local_extra)

            yield local_extra['on_ready']

            # connect the remote session
            #
            # remote connection parameters to ApplicationRunner:
            #
            # url: The WebSocket URL of the WAMP router to connect to (e.g. ws://somehost.com:8090/somepath)
            # realm: The WAMP realm to join the application session to.
            # extra: Optional extra configuration to forward to the application component.
            # serializers: List of :class:`autobahn.wamp.interfaces.ISerializer` (or None for default serializers).
            # ssl: None or :class:`twisted.internet.ssl.CertificateOptions`
            # proxy: Explicit proxy server to use; a dict with ``host`` and ``port`` keys
            # headers: Additional headers to send (only applies to WAMP-over-WebSocket).
            # max_retries: Maximum number of reconnection attempts. Unlimited if set to -1.
            # initial_retry_delay: Initial delay for reconnection attempt in seconds (Default: 1.0s).
            # max_retry_delay: Maximum delay for reconnection attempts in seconds (Default: 60s).
            # retry_delay_growth: The growth factor applied to the retry delay between reconnection attempts (Default 1.5).
            # retry_delay_jitter: A 0-argument callable that introduces nose into the delay. (Default random.random)
            #
            remote_runner = ApplicationRunner(url=link_config.transport['url'], realm=remote_realm, extra=remote_extra)

            yield remote_runner.run(remote_session,
                                    start_reactor=False,
                                    auto_reconnect=True,
                                    endpoint=connecting_endpoint,
                                    reactor=self._controller._reactor)

            yield remote_extra['on_ready']

        except:
            # make sure to remove the half-initialized link from our map ..
            del self._links[link_id]

            # .. and then re-raise
            raise

        # the router link is established: store final infos
        rlink.started = time_ns()
        rlink.started_by = caller
        rlink.local = local_session
        rlink.remote = remote_session

        return rlink

    @inlineCallbacks
    def stop_link(self, link_id, caller):
        raise NotImplementedError()



###############################################################################
#
# Crossbar.io Shell
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import sys
from pprint import pprint
# import iso8601
import numpy as np

# http://babel.pocoo.org/en/latest/dates.html
# from babel.dates import get_timezone

import txaio

txaio.use_twisted()

from txaio import time_ns, sleep

from twisted.internet.task import react
from twisted.internet.defer import inlineCallbacks

from crossbar.shell.client import create_management_session
from crossbar.master import personality

try:
    import curses
    import curses.panel
    from curses import wrapper
except ImportError:
    print('fatal: curses module not found')
    sys.exit(1)


@inlineCallbacks
def twisted_main(reactor, stdscr=None, mrealms=None, management_url=None, privkey_file=None):
    mrealms = mrealms or ['default']
    try:
        if stdscr:
            stdscr.clear()
            y = 5
            for line in personality.Personality.BANNER.splitlines():
                stdscr.addstr(y, 20, line, curses.color_pair(227))
                y += 1
            y += 3
            stdscr.addstr(y, 24, 'Please wait while collecting data from managed nodes ...')
            stdscr.refresh()

        while True:
            if stdscr:
                stdscr.clear()
                y = 0

                stdscr.addstr(y, 0, '=' * 240)
                y += 1

                x = 0
                stdscr.addstr(y, x, 'Node')
                x += 34
                stdscr.addstr(y, x, 'Mgmt Realm', curses.color_pair(14))
                x += 15
                stdscr.addstr(y, x, 'Node ID', curses.color_pair(14))
                x += 25
                stdscr.addstr(y, x, 'Node OID', curses.color_pair(14))
                x += 40
                stdscr.addstr(y, x, 'Status')
                x += 10
                stdscr.addstr(y, x, 'Last Heartbeat')
                x += 12

                x += 4
                stdscr.addstr(y, x + 0, ' Usr')
                stdscr.addstr(y, x + 5, ' Sys')
                stdscr.addstr(y, x + 10, ' Idl')

                x += 1
                stdscr.addstr(y, x + 15, ' Mem')
                stdscr.addstr(y, x + 20, ' IPv4 sckts')
                x += 5 * 5

                x += 11
                stdscr.addstr(y, x + 0, ' Pxy', curses.color_pair(41))
                stdscr.addstr(y, x + 4, ' Rtr', curses.color_pair(41))
                stdscr.addstr(y, x + 8, ' Xbr', curses.color_pair(41))
                stdscr.addstr(y, x + 12, ' Cnt', curses.color_pair(41))
                stdscr.addstr(y, x + 16, ' Gst', curses.color_pair(41))
                x += 4 * 5

                x += 4
                stdscr.addstr(y, x + 0, ' Rlm', curses.color_pair(227))
                stdscr.addstr(y, x + 5, ' Rls')
                stdscr.addstr(y, x + 10, ' Rlk')
                stdscr.addstr(y, x + 15, '  Sessions', curses.color_pair(41))
                stdscr.addstr(y, x + 25, '  Messages', curses.color_pair(41))
                x += 5 * 5

                y += 1
                stdscr.addstr(y, 0, '-' * 240)
                y += 1

            last_mrealm = None

            for mrealm in mrealms:
                if stdscr:
                    if last_mrealm:
                        stdscr.addstr(y, 0, '.' * 240)
                        y += 1

                try:
                    session, _ = yield create_management_session(url=management_url,
                                                                 realm=mrealm,
                                                                 privkey_file=privkey_file)
                except Exception as e:
                    print(e)

                node_oids = yield session.call('crossbarfabriccenter.mrealm.get_nodes')
                if not stdscr:
                    pprint(node_oids)
                for node_oid in node_oids:
                    node = yield session.call('crossbarfabriccenter.mrealm.get_node', node_oid)
                    if not stdscr:
                        pprint(node)

                    if True:
                        node_authid = node['authid']
                        # node_pubkey = node['pubkey']
                        if node and node['status'] == 'online':

                            node_status = yield session.call('crossbarfabriccenter.remote.node.get_status', node_oid)

                            if not stdscr:
                                pprint(node_status)

                            node_system_stats = yield session.call('crossbarfabriccenter.remote.node.get_system_stats',
                                                                   node_oid)

                            if not stdscr:
                                pprint(node_system_stats)

                            cpu_user = node_system_stats['cpu']['user']
                            cpu_system = node_system_stats['cpu']['system']
                            cpu_idle = node_system_stats['cpu']['idle']
                            # memory_total = node_system_stats['memory']['total']
                            # memory_avail = node_system_stats['memory']['available']
                            memory_perc = node_system_stats['memory']['percent']
                            # network_recv = node_system_stats['network']['bytes_recv_per_sec']
                            # network_sent = node_system_stats['network']['bytes_sent_per_sec']
                            network_conns = node_system_stats['network']['connection']['AF_INET']

                            # tz = get_timezone('Europe/Berlin')
                            # started = format_datetime(iso8601.parse_date(node_status['started']), tzinfo=tz, locale='de_DE', format='long')

                            last_heartbeat = np.datetime64(node['timestamp'], 'ns')
                            now = np.datetime64(time_ns(), 'ns')
                            if now > last_heartbeat:
                                last_heartbeat_ago = str((now - last_heartbeat).astype("timedelta64[s]"))
                            else:
                                last_heartbeat_ago = None
                            node_title = node_status['title']

                            # get IDs for all workers running in this node
                            worker_info = {}
                            router_info = {}
                            workers = yield session.call('crossbarfabriccenter.remote.node.get_workers', node_oid)
                            for worker_id in workers:
                                # get worker detail information
                                # {'id': 'xbr1', 'pid': 11507, 'type': 'marketplace', 'status': 'started', 'created': '2020-06-22T06:15:44.589Z', 'started': '2020-06-22T06:15:48.224Z', 'startup_time': 3.635574, 'uptime': 13949.814363}
                                worker = yield session.call('crossbarfabriccenter.remote.node.get_worker', node_oid,
                                                            worker_id)
                                if not stdscr:
                                    pprint(worker)

                                if worker['status'] == 'started':
                                    if worker['type'] not in worker_info:
                                        worker_info[worker['type']] = 0
                                    worker_info[worker['type']] += 1

                                    if worker['type'] == 'router':
                                        # get IDs for all realm running in router worker
                                        realm_oids = yield session.call(
                                            'crossbarfabriccenter.remote.router.get_router_realms', node_oid,
                                            worker_id)
                                        for realm_oid in realm_oids:
                                            # get realm detail information
                                            realm = yield session.call(
                                                'crossbarfabriccenter.remote.router.get_router_realm', node_oid,
                                                worker_id, realm_oid)

                                            if not stdscr:
                                                pprint(realm)

                                            # get per-realm messaging statistics
                                            realm_stats = yield session.call(
                                                'crossbarfabriccenter.remote.router.get_router_realm_stats', node_oid,
                                                worker_id, realm_oid)

                                            if not stdscr:
                                                pprint(realm_stats)

                                            realm_id = realm['id']
                                            realm_name = realm['config']['name']
                                            realm_created = realm['created']

                                            ri_obj = {
                                                'node_oid': node_oid,
                                                'worker_id': worker_id,
                                                'id': realm_id,
                                                'name': realm_name,
                                                'created': realm_created,
                                                'rlinks': len([1 for rlink in realm['rlinks'] if rlink['connected']]),
                                            }

                                            sw_latest = '20.6.2.dev2' in node_status['title']
                                            if sw_latest:
                                                # get IDs of all rlinks running in this router worker and realm
                                                rlink_oids = yield session.call(
                                                    'crossbarfabriccenter.remote.router.get_router_realm_links',
                                                    node_oid, worker_id, realm_oid)
                                                ri_obj['rlinks'] = len(rlink_oids)

                                            # {'realm001': {'messages': {'received': {'publish': 39, 'register': 42},
                                            #                            'sent': {'registered': 42}},
                                            #               'roles': 4,
                                            #               'sessions': 2}}
                                            received = realm_stats[realm_id]['messages']['received']
                                            sent = realm_stats[realm_id]['messages']['sent']
                                            total = 0
                                            for k in received:
                                                total += received[k]
                                            for k in sent:
                                                total += sent[k]

                                            ri_obj['messages'] = total
                                            ri_obj['received'] = realm_stats[realm_id]['messages']['received']
                                            ri_obj['sent'] = realm_stats[realm_id]['messages']['sent']

                                            ri_obj['sessions'] = realm_stats[realm_id]['sessions']
                                            ri_obj['roles'] = realm_stats[realm_id]['roles']

                                            if realm_name not in router_info:
                                                router_info[realm_name] = [ri_obj]
                                            else:
                                                router_info[realm_name].append(ri_obj)

                        else:
                            worker_info = {}
                            router_info = {}
                            # started = '-'
                            last_heartbeat_ago = '-'
                            node_title = '-'
                            cpu_user = 0
                            cpu_system = 0
                            cpu_idle = 0
                            # memory_total = 0
                            # memory_avail = 0
                            memory_perc = 0
                            # network_recv = 0
                            # network_sent = 0
                            network_conns = 0

                        if stdscr:

                            def fmt(data, key):
                                val = data.get(key, 0)
                                if val:
                                    return '{0: >4}'.format(val), curses.color_pair(41)
                                else:
                                    return '   -', curses.color_pair(8)

                            x = 0

                            stdscr.addstr(y, x, node_title)
                            x += 34

                            stdscr.addstr(y, x, mrealm, curses.color_pair(14))
                            x += 15

                            stdscr.addstr(y, x, node_authid, curses.color_pair(14))
                            x += 25

                            stdscr.addstr(y, x, node_oid, curses.color_pair(14))
                            x += 40

                            if node['status'] == 'online':
                                stdscr.addstr(y, x, node['status'], curses.color_pair(41))
                            else:
                                stdscr.addstr(y, x, node['status'], curses.color_pair(10))
                            x += 10

                            stdscr.addstr(y, x, last_heartbeat_ago)
                            x += 12

                            def fmt2(val):
                                return '{0: >4}'.format(val), curses.color_pair(8)

                            x += 4
                            stdscr.addstr(y, x + 0, *fmt2(round(cpu_user, 1)))
                            stdscr.addstr(y, x + 5, *fmt2(round(cpu_system, 1)))
                            stdscr.addstr(y, x + 10, *fmt2(round(cpu_idle, 1)))

                            x += 1
                            stdscr.addstr(y, x + 15, *fmt2(round(memory_perc, 1)))

                            x += 1
                            stdscr.addstr(y, x + 20, '{0: >10}'.format(network_conns))

                            x += 5 * 5

                            x += 10
                            stdscr.addstr(y, x + 0, *fmt(worker_info, 'proxy'))
                            stdscr.addstr(y, x + 4, *fmt(worker_info, 'router'))
                            stdscr.addstr(y, x + 8, *fmt(worker_info, 'marketplace'))
                            stdscr.addstr(y, x + 12, *fmt(worker_info, 'container'))
                            stdscr.addstr(y, x + 16, *fmt(worker_info, 'guest'))
                            x += 4 * 5

                        # cpu_user = node_system_stats['cpu']['user']
                        # cpu_system = node_system_stats['cpu']['system']
                        # cpu_idle = node_system_stats['cpu']['idle']
                        # memory_total = node_system_stats['memory']['total']
                        # memory_avail = node_system_stats['memory']['available']
                        # memory_perc = node_system_stats['memory']['percent']
                        # network_recv = node_system_stats['network']['bytes_recv_per_sec']
                        # network_sent = node_system_stats['network']['bytes_sent_per_sec']
                        # network_conns = node_system_stats['network']['connection']['AF_INET']

                        roles = 0
                        sessions = 0
                        messages = 0
                        rlinks = 0
                        for realm_id in router_info:
                            for realm_obj in router_info[realm_id]:
                                roles += realm_obj['roles']
                                sessions += realm_obj['sessions']
                                messages += realm_obj['messages']
                                rlinks += realm_obj['rlinks']

                        if stdscr:
                            x += 4
                            stdscr.addstr(y, x + 0, '{0: >4}'.format(len(router_info.keys())), curses.color_pair(227))
                            stdscr.addstr(y, x + 5, '{0: >4}'.format(roles))
                            stdscr.addstr(y, x + 10, '{0: >4}'.format(rlinks))
                            stdscr.addstr(y, x + 15, '{0: >10}'.format(sessions), curses.color_pair(41))
                            stdscr.addstr(y, x + 25, '{0: >10}'.format(messages), curses.color_pair(41))
                            x += 5 * 5

                            y += 1

                last_mrealm = mrealm

            if stdscr:
                stdscr.addstr(y, 0, '=' * 240)
                y += 1

                stdscr.refresh()

            yield sleep(5)
    except Exception as e:
        sys.stderr.write(str(e))
        sys.exit(1)


def main(stdscr=None, mrealms=None, management_url=None, privkey_file=None):
    if stdscr:
        # https://stackoverflow.com/a/22166613/884770
        curses.start_color()
        curses.use_default_colors()
        curses.curs_set(0)
        for i in range(0, curses.COLORS):
            curses.init_pair(i + 1, i, -1)

    react(twisted_main, [stdscr, mrealms or ['default'], management_url, privkey_file])


def run(management_url=None, privkey_file=None):
    try:
        wrapper(main, management_url=management_url, privkey_file=privkey_file)
        # main(management_url=management_url, privkey_file=privkey_file)
    except Exception as e:
        print(e)


if __name__ == '__main__':
    run()



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import txaio

from crossbar.network.node.node import XbrNetworkNode
from crossbar.edge.personality import Personality as _Personality


class Personality(_Personality):

    log = txaio.make_logger()

    NAME = 'network'

    Node = XbrNetworkNode

    def check_config(self, personality):
        res = _Personality.check_config(self, personality)
        return res

    def check_config_file(self, personality, config_path):
        res = _Personality.check_config_file(self, personality, config_path)
        return res



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

from crossbar.network.node.node import XbrNetworkNode, XbrNetworkNodeControllerSession

__all__ = ('XbrNetworkNode', 'XbrNetworkNodeControllerSession')



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from crossbar.master.api.remote import RemoteApi

__all__ = ('RemoteContainerApi', )


class RemoteContainerApi(RemoteApi):

    # remote procedures and topics are exposed under this prefix
    PREFIX = u'crossbarfabriccenter.remote.container.'

    PROCS = {
        # these are worker level procedures
        u'worker': [
            (u'get_components', u'get_components'),
            (u'get_component', u'get_component'),
            (u'start_component', u'start_component'),
            (u'stop_component', u'stop_component'),
        ],
    }

    EVENTS = {
        # these are worker level topics
        u'worker': [
            u'on_container_component_starting',
            u'on_container_component_started',
            u'on_container_component_stopping',
            u'on_container_component_stopped',
        ]
    }



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import os
import crossbar
import binascii

from twisted import internet
# from twisted.protocols.tls import TLSMemoryBIOProtocol

from autobahn.twisted import websocket
from autobahn.twisted import rawsocket
from autobahn.websocket.compress import PerMessageDeflateOffer, PerMessageDeflateOfferAccept

from autobahn.wamp.types import TransportDetails
from autobahn.websocket.types import ConnectionDeny
from autobahn.util import hlval, hltype, hlid

from txaio import make_logger

from crossbar.router.cookiestore import CookieStoreMemoryBacked, CookieStoreFileBacked, CookieStoreDatabaseBacked

from crossbar.common.twisted.endpoint import create_connecting_endpoint_from_config

log = make_logger()

__all__ = (
    'WampWebSocketServerFactory',
    'WampRawSocketServerFactory',
    'WampWebSocketServerProtocol',
    'WampRawSocketServerProtocol',
    'WampWebSocketClientFactory',
    'WampRawSocketClientFactory',
    'WampWebSocketClientProtocol',
    'WampRawSocketClientProtocol',
)


def set_websocket_options(factory, options):
    """
    Set WebSocket options on a WebSocket or WAMP-WebSocket factory.

    :param factory: The WebSocket or WAMP-WebSocket factory to set options on.
    :type factory:  Instance of :class:`autobahn.twisted.websocket.WampWebSocketServerFactory`
                    or :class:`autobahn.twisted.websocket.WebSocketServerFactory`.
    :param options: Options from Crossbar.io transport configuration.
    :type options: dict
    """
    c = options

    # we need to pop() this, since it is not a WebSocket option to be consumed
    # by setProtocolOption(), but will get used in onConnect() ("STRICT_PROTOCOL_NEGOTIATION")
    #
    factory._requireWebSocketSubprotocol = c.pop("require_websocket_subprotocol", True)

    versions = []
    if c.get("enable_hybi10", True):
        versions.append(8)
    if c.get("enable_rfc6455", True):
        versions.append(13)

    # FIXME: enforce!!
    #
    # self.connectionCap = c.get("max_connections")

    # convert to seconds
    #
    openHandshakeTimeout = float(c.get("open_handshake_timeout", 0))
    if openHandshakeTimeout:
        openHandshakeTimeout = openHandshakeTimeout / 1000.

    closeHandshakeTimeout = float(c.get("close_handshake_timeout", 0))
    if closeHandshakeTimeout:
        closeHandshakeTimeout = closeHandshakeTimeout / 1000.

    autoPingInterval = None
    if "auto_ping_interval" in c:
        autoPingInterval = float(c["auto_ping_interval"]) / 1000.

    autoPingTimeout = None
    if "auto_ping_timeout" in c:
        autoPingTimeout = float(c["auto_ping_timeout"]) / 1000.

    # WebSocket compression
    #
    per_msg_compression = lambda _: None  # noqa
    if 'compression' in c:

        # permessage-deflate
        #
        if 'deflate' in c['compression']:

            log.debug("enabling WebSocket compression (permessage-deflate)")

            params = c['compression']['deflate']

            request_no_context_takeover = params.get('request_no_context_takeover', False)
            request_max_window_bits = params.get('request_max_window_bits', 0)
            no_context_takeover = params.get('no_context_takeover', None)
            window_bits = params.get('max_window_bits', None)
            mem_level = params.get('memory_level', None)

            def accept(offers):
                for offer in offers:
                    if isinstance(offer, PerMessageDeflateOffer):
                        if (request_max_window_bits == 0 or offer.accept_max_window_bits) and \
                           (not request_no_context_takeover or offer.accept_no_context_takeover):
                            return PerMessageDeflateOfferAccept(
                                offer,
                                request_max_window_bits=request_max_window_bits,
                                request_no_context_takeover=request_no_context_takeover,
                                no_context_takeover=no_context_takeover,
                                window_bits=window_bits,
                                mem_level=mem_level)

            per_msg_compression = accept

    if factory.isServer:
        factory.setProtocolOptions(
            versions=versions,
            webStatus=c.get("enable_webstatus", True),
            utf8validateIncoming=c.get("validate_utf8", True),
            maskServerFrames=c.get("mask_server_frames", False),
            requireMaskedClientFrames=c.get("require_masked_client_frames", True),
            applyMask=c.get("apply_mask", True),
            maxFramePayloadSize=c.get("max_frame_size", 0),
            maxMessagePayloadSize=c.get("max_message_size", 0),
            autoFragmentSize=c.get("auto_fragment_size", 0),
            failByDrop=c.get("fail_by_drop", False),
            echoCloseCodeReason=c.get("echo_close_codereason", False),
            openHandshakeTimeout=openHandshakeTimeout,
            closeHandshakeTimeout=closeHandshakeTimeout,
            tcpNoDelay=c.get("tcp_nodelay", True),
            autoPingInterval=autoPingInterval,
            autoPingTimeout=autoPingTimeout,
            autoPingSize=c.get("auto_ping_size", None),
            autoPingRestartOnAnyTraffic=c.get("auto_ping_restart_on_any_traffic", None),
            serveFlashSocketPolicy=c.get("enable_flash_policy", None),
            flashSocketPolicy=c.get("flash_policy", None),
            allowedOrigins=c.get("allowed_origins", ["*"]),
            allowNullOrigin=bool(c.get("allow_null_origin", True)),
            perMessageCompressionAccept=per_msg_compression,
        )
    else:
        factory.setProtocolOptions(
            version=None,
            utf8validateIncoming=c.get("validate_utf8", True),
            acceptMaskedServerFrames=c.get("accept_masked_server_frames", False),
            maskClientFrames=c.get("mask_client_frames", True),
            applyMask=c.get("apply_mask", True),
            maxFramePayloadSize=c.get("max_frame_size", 0),
            maxMessagePayloadSize=c.get("max_message_size", 0),
            autoFragmentSize=c.get("auto_fragment_size", 0),
            failByDrop=c.get("fail_by_drop", False),
            echoCloseCodeReason=c.get("echo_close_codereason", False),
            openHandshakeTimeout=openHandshakeTimeout,
            closeHandshakeTimeout=closeHandshakeTimeout,
            tcpNoDelay=c.get("tcp_nodelay", True),
            autoPingInterval=autoPingInterval,
            autoPingTimeout=autoPingTimeout,
            autoPingSize=c.get("auto_ping_size", None),
            autoPingRestartOnAnyTraffic=c.get("auto_ping_restart_on_any_traffic", None),
            perMessageCompressionOffers=None,
            perMessageCompressionAccept=None,
        )


class WampWebSocketServerProtocol(websocket.WampWebSocketServerProtocol):
    """
    Crossbar.io WAMP-over-WebSocket server protocol.
    """
    log = make_logger()

    def __init__(self):
        super(WampWebSocketServerProtocol, self).__init__()
        self._cbtid = None

    def onConnect(self, request):

        self.log.debug('{func}(request={request})', func=hltype(self.onConnect), request=request)

        if self.factory.debug_traffic:
            from twisted.internet import reactor

            def print_traffic():
                self.log.info(
                    "Traffic {peer}: {wire_in} / {wire_out} in / out bytes - {ws_in} / {ws_out} in / out msgs",
                    peer=self.peer,
                    wire_in=self.trafficStats.incomingOctetsWireLevel,
                    wire_out=self.trafficStats.outgoingOctetsWireLevel,
                    ws_in=self.trafficStats.incomingWebSocketMessages,
                    ws_out=self.trafficStats.outgoingWebSocketMessages,
                )
                reactor.callLater(1, print_traffic)

            print_traffic()

        # if WebSocket client did not set WS subprotocol, assume "wamp.2.json"
        #
        self.STRICT_PROTOCOL_NEGOTIATION = self.factory._requireWebSocketSubprotocol

        # handle WebSocket opening handshake
        #
        protocol, headers = websocket.WampWebSocketServerProtocol.onConnect(self, request)

        self.log.debug('{func}: proceed with WebSocket opening handshake for WebSocket subprotocol "{protocol}"',
                       func=hltype(self.onConnect),
                       protocol=hlval(protocol))

        try:

            self._origin = request.origin

            # transport-level WMAP authentication info
            #
            self._authid = None
            self._authrole = None
            self._authrealm = None
            self._authmethod = None
            self._authextra = None
            self._authprovider = None

            # cookie tracking and cookie-based authentication
            #
            self._cbtid = None

            if self.factory._cookiestore:

                # try to parse an already set cookie from HTTP request headers
                self._cbtid = self.factory._cookiestore.parse(request.headers)

                if self._cbtid:
                    self.log.info(
                        '{func}: parsed tracking/authentication cookie cbtid "{cbtid}" from HTTP request headers',
                        func=hltype(self.onConnect),
                        cbtid=hlval(self._cbtid))
                else:
                    self.log.info('{func}: no tracking/authentication cookie cbtid found in HTTP request headers!',
                                  func=hltype(self.onConnect))

                # if no cookie is set, or it doesn't exist in our database, create a new cookie
                if self._cbtid is None or not self.factory._cookiestore.exists(self._cbtid):

                    self._cbtid, headers['Set-Cookie'] = self.factory._cookiestore.create()

                    if 'cookie' in self.factory._config:
                        if 'secure' in self.factory._config[
                                'cookie'] and self.factory._config['cookie']['secure'] is True:
                            headers['Set-Cookie'] += ';Secure'
                        if 'http_strict' in self.factory._config[
                                'cookie'] and self.factory._config['cookie']['http_strict'] is True:
                            headers['Set-Cookie'] += ';HttpOnly'
                        if 'same_site' in self.factory._config['cookie']:
                            headers['Set-Cookie'] += ';SameSite=' + self.factory._config['cookie']['same_site']

                    self.log.info('{func}: setting new cookie {cookie}',
                                  func=hltype(self.onConnect),
                                  cookie=hlval(headers['Set-Cookie']))
                else:
                    self.log.info('{func}: tracking/authentication cookie cbtid "{cbtid}" already set and stored',
                                  func=hltype(self.onConnect),
                                  cbtid=hlval(self._cbtid))

                # add this WebSocket connection to the set of connections
                # associated with the same cookie
                self.factory._cookiestore.addProto(self._cbtid, self)

                self.log.debug("Cookie tracking enabled on WebSocket connection {ws}", ws=self)

                # if cookie-based authentication is enabled, set auth info from cookie store
                #
                if 'auth' in self.factory._config and 'cookie' in self.factory._config['auth']:

                    self._authid, self._authrole, self._authmethod, self._authrealm, self._authextra = self.factory._cookiestore.getAuth(
                        self._cbtid)

                    if self._authid:
                        # there is a cookie set, and the cookie was previously successfully authenticated,
                        # so immediately authenticate the client using that information
                        self._authprovider = 'cookie'
                        self.log.info(
                            '{func} authenticated client via cookie {cookiename}={cbtid} as authid="{authid}", authrole="{authrole}", authmethod="{authmethod}", authprovider="{authprovider}", authrealm="{authrealm}"',
                            func=hltype(self.onConnect),
                            cookiename=self.factory._cookiestore._cookie_id_field,
                            cbtid=hlid(self._cbtid),
                            authid=hlid(self._authid),
                            authrole=hlid(self._authrole),
                            authmethod=hlval(self._authmethod),
                            authprovider=hlval(self._authprovider),
                            authrealm=hlid(self._authrealm))
                    else:
                        # there is a cookie set, but the cookie wasn't authenticated yet using a different auth method
                        self.log.info(
                            '{func} cookie-based authentication enabled, but cookie is not authenticated yet',
                            func=hltype(self.onConnect))
                else:
                    self.log.info('{func} cookie-based authentication disabled on connection',
                                  func=hltype(self.onConnect))
            else:
                self.log.info('{func} cookie tracking disabled on WebSocket connection', func=hltype(self.onConnect))

            # negotiated WebSocket subprotocol in use, e.g. "wamp.2.cbor.batched"
            self._transport_details.websocket_protocol = protocol

            # WebSocket extensions in use. will be filled in onOpen(), see below
            self._transport_details.websocket_extensions_in_use = None

            # Crossbar.io tracking ID (for cookie tracking)
            self._transport_details.http_cbtid = self._cbtid

            # all HTTP headers as received by the WebSocket client
            self._transport_details.http_headers_received = request.headers

            # only customer user headers (such as cookie)
            self._transport_details.http_headers_sent = headers

            # accept the WebSocket connection, speaking subprotocol `protocol`
            # and setting HTTP headers `headers`
            return protocol, headers

        except:
            self.log.failure()

    def onOpen(self):
        # note the WebSocket extensions negotiated
        self._transport_details.websocket_extensions_in_use = [e.__json__() for e in self.websocket_extensions_in_use]

        return super(WampWebSocketServerProtocol, self).onOpen()

    def sendServerStatus(self, redirectUrl=None, redirectAfter=0):
        """
        Used to send out server status/version upon receiving a HTTP/GET without
        upgrade to WebSocket header (and option serverStatus is True).
        """
        try:
            page = self.factory._templates.get_template('cb_ws_status.html')
            self.sendHtml(
                page.render(redirectUrl=redirectUrl,
                            redirectAfter=redirectAfter,
                            cbVersion=crossbar.__version__,
                            wsUri=self.factory.url,
                            peer=self.peer,
                            workerPid=os.getpid()))
        except Exception:
            self.log.failure("Error rendering WebSocket status page template: {log_failure.value}")

    def onClose(self, wasClean, code, reason):
        super(WampWebSocketServerProtocol, self).onClose(wasClean, code, reason)

        # remove this WebSocket connection from the set of connections
        # associated with the same cookie
        if self._cbtid:
            self.factory._cookiestore.dropProto(self._cbtid, self)


class WampWebSocketServerFactory(websocket.WampWebSocketServerFactory):
    """
    Crossbar.io WAMP-over-WebSocket server factory.
    """

    showServerVersion = False
    protocol = WampWebSocketServerProtocol
    log = make_logger()

    def __init__(self, factory, cbdir, config, templates):
        """

        :param factory: WAMP session factory.
        :type factory: An instance of ..

        :param cbdir: The Crossbar.io node directory.
        :type cbdir: str

        :param config: Crossbar transport configuration.
        :type config: dict

        :param templates:
        :type templates:
        """
        self.debug_traffic = config.get('debug_traffic', False)

        options = config.get('options', {})

        # announce Crossbar.io server version
        #
        self.showServerVersion = options.get('show_server_version', self.showServerVersion)
        if self.showServerVersion:
            server = "Crossbar/{}".format(crossbar.__version__)
        else:
            # do not disclose crossbar version
            server = "Crossbar"

        # external (public) listening port (eg when running behind a reverse proxy)
        #
        externalPort = options.get('external_port', None)

        # explicit list of WAMP serializers
        #
        if 'serializers' in config:
            serializers = []
            sers = set(config['serializers'])

            if 'flatbuffers' in sers:
                # try FlatBuffers WAMP serializer
                try:
                    from autobahn.wamp.serializer import FlatBuffersSerializer
                    serializers.append(FlatBuffersSerializer(batched=True))
                    serializers.append(FlatBuffersSerializer())
                except ImportError('FlatBuffersSerializer'):
                    self.log.warn("Warning: could not load WAMP-FlatBuffers serializer")
                else:
                    sers.discard('flatbuffers')

            if 'cbor' in sers:
                # try CBOR WAMP serializer
                try:
                    from autobahn.wamp.serializer import CBORSerializer
                    serializers.append(CBORSerializer(batched=True))
                    serializers.append(CBORSerializer())
                except ImportError('CBORSerializer'):
                    self.log.warn("Warning: could not load WAMP-CBOR serializer")
                else:
                    sers.discard('cbor')

            if 'msgpack' in sers:
                # try MsgPack WAMP serializer
                try:
                    from autobahn.wamp.serializer import MsgPackSerializer
                    serializers.append(MsgPackSerializer(batched=True))
                    serializers.append(MsgPackSerializer())
                except ImportError('MsgPackSerializer'):
                    self.log.warn("Warning: could not load WAMP-MsgPack serializer")
                else:
                    sers.discard('msgpack')

            if 'ubjson' in sers:
                # try UBJSON WAMP serializer
                try:
                    from autobahn.wamp.serializer import UBJSONSerializer
                    serializers.append(UBJSONSerializer(batched=True))
                    serializers.append(UBJSONSerializer())
                except ImportError('UBJSONSerializer'):
                    self.log.warn("Warning: could not load WAMP-UBJSON serializer")
                else:
                    sers.discard('ubjson')

            if 'json' in sers:
                # try JSON WAMP serializer
                try:
                    from autobahn.wamp.serializer import JsonSerializer
                    serializers.append(JsonSerializer(batched=True))
                    serializers.append(JsonSerializer())
                except ImportError('JsonSerializer'):
                    self.log.warn("Warning: could not load WAMP-JSON serializer")
                else:
                    sers.discard('json')

            if not serializers:
                raise Exception("no valid WAMP serializers specified")

            if len(sers) > 0:
                raise Exception("invalid WAMP serializers specified (the following were unprocessed) {}".format(sers))

        else:
            serializers = None

        websocket.WampWebSocketServerFactory.__init__(self,
                                                      factory,
                                                      serializers=serializers,
                                                      url=config.get('url', None),
                                                      server=server,
                                                      externalPort=externalPort)

        # Crossbar.io node directory
        self._cbdir = cbdir

        # transport configuration
        self._config = config

        # Jinja2 templates for 404 etc
        self._templates = templates

        # enable cookie tracking if a cookie store is configured
        if 'cookie' in config:
            # cookie store configuration item
            cookie_config = config['cookie']

            # cookie store
            cookie_store_config = cookie_config['store']
            cookie_store_type = cookie_store_config['type']

            # setup ephemeral, memory-backed cookie store
            if cookie_store_type == 'memory':
                self._cookiestore = CookieStoreMemoryBacked(cookie_config)
                self.log.info("Memory-backed cookie store active.")

            # setup persistent, file-backed cookie store
            elif cookie_store_type == 'file':
                cookie_store_file = os.path.abspath(os.path.join(self._cbdir, cookie_store_config['filename']))
                self._cookiestore = CookieStoreFileBacked(cookie_store_file, cookie_config)
                self.log.info("File-backed cookie store active {cookie_store_file}",
                              cookie_store_file=hlval(cookie_store_file))

            # setup persistent, database-backed cookie store
            elif cookie_store_type == 'database':
                cookie_dbpath = os.path.abspath(os.path.join(self._cbdir, cookie_store_config['path']))
                self._cookiestore = CookieStoreDatabaseBacked(cookie_dbpath, cookie_config)
                self.log.info("Database-backed cookie store active! [cookiestore={cookiestore}]",
                              cookiestore=hltype(CookieStoreDatabaseBacked))

            else:
                # should not arrive here as the config should have been checked before
                raise NotImplementedError('{}: implementation of cookiestore of type "{}" missing'.format(
                    self.__class__.__name__, cookie_store_type))
        else:
            # this disables cookie tracking (both with or without WAMP-cookie authentication)
            self._cookiestore = None

        # set WebSocket options
        set_websocket_options(self, options)


def set_rawsocket_options(factory, options):
    """
    Set RawSocket options on a RawSocket or WAMP-RawSocket (server or client )factory.

    :param factory: The RawSocket or WAMP-RawSocket factory to set options on.
    :type factory:  Instance of :class:`autobahn.twisted.rawsocket.WampRawSocketServerFactory`
                    or :class:`autobahn.twisted.rawsocket.WampRawSocketClientFactory`.
    :param options: RawSocket transport options item from Crossbar.io transport configuration.
    :type options: dict
    """
    c = options
    factory.setProtocolOptions(maxMessagePayloadSize=c.get("max_message_size", None))


class WampRawSocketServerProtocol(rawsocket.WampRawSocketServerProtocol):
    """
    Crossbar.io WAMP-over-RawSocket server protocol.
    """
    log = make_logger()

    def connectionMade(self):
        rawsocket.WampRawSocketServerProtocol.connectionMade(self)

        # transport authentication
        #
        self._authid = None
        self._authrole = None
        self._authrealm = None
        self._authmethod = None
        self._authprovider = None
        self._authextra = None

        # cookie tracking ID
        #
        self._cbtid = None

    def _on_handshake_complete(self):
        self._transport_details.channel_serializer = TransportDetails.CHANNEL_SERIALIZER_FROM_STR[
            self._serializer.SERIALIZER_ID]
        self._transport_details.websocket_protocol = 'wamp.2.{}'.format(self._serializer.SERIALIZER_ID)
        return rawsocket.WampRawSocketServerProtocol._on_handshake_complete(self)


class WampRawSocketServerFactory(rawsocket.WampRawSocketServerFactory):
    """
    Crossbar.io WAMP-over-RawSocket server factory.
    """

    protocol = WampRawSocketServerProtocol
    log = make_logger()

    def __init__(self, factory, config):

        # remember transport configuration
        #
        self._config = config

        # explicit list of WAMP serializers
        #
        if 'serializers' in config:
            serializers = []
            sers = set(config['serializers'])

            if 'flatbuffers' in sers:
                # try FlatBuffers WAMP serializer
                try:
                    from autobahn.wamp.serializer import FlatBuffersSerializer
                    serializers.append(FlatBuffersSerializer())
                except ImportError:
                    self.log.warn("Warning: could not load WAMP-FlatBuffers serializer")
                else:
                    sers.discard('flatbuffers')

            if 'cbor' in sers:
                # try CBOR WAMP serializer
                try:
                    from autobahn.wamp.serializer import CBORSerializer
                    serializers.append(CBORSerializer())
                except ImportError:
                    self.log.warn("Warning: could not load WAMP-CBOR serializer")
                else:
                    sers.discard('cbor')

            if 'msgpack' in sers:
                # try MsgPack WAMP serializer
                try:
                    from autobahn.wamp.serializer import MsgPackSerializer
                    serializer = MsgPackSerializer()
                    serializer._serializer.ENABLE_V5 = False  # FIXME
                    serializers.append(serializer)
                except ImportError:
                    self.log.warn("Warning: could not load WAMP-MsgPack serializer")
                else:
                    sers.discard('msgpack')

            if 'ubjson' in sers:
                # try UBJSON WAMP serializer
                try:
                    from autobahn.wamp.serializer import UBJSONSerializer
                    serializers.append(UBJSONSerializer(batched=True))
                    serializers.append(UBJSONSerializer())
                except ImportError:
                    self.log.warn("Warning: could not load WAMP-UBJSON serializer")
                else:
                    sers.discard('ubjson')

            if 'json' in sers:
                # try JSON WAMP serializer
                try:
                    from autobahn.wamp.serializer import JsonSerializer
                    serializers.append(JsonSerializer())
                except ImportError:
                    self.log.warn("Warning: could not load WAMP-JSON serializer")
                else:
                    sers.discard('json')

            if not serializers:
                raise Exception("no valid WAMP serializers specified")

            if len(sers) > 0:
                raise Exception("invalid WAMP serializers specified (the following were unprocessed) {}".format(sers))

        else:
            serializers = None

        rawsocket.WampRawSocketServerFactory.__init__(self, factory, serializers)

        if 'options' in config:
            set_rawsocket_options(self, config['options'])

        self.log.debug(
            "RawSocket transport factory created using {serializers} serializers, max. message size {maxsize}",
            serializers=serializers,
            maxsize=self._max_message_size)


class WampWebSocketClientProtocol(websocket.WampWebSocketClientProtocol):
    """
    Crossbar.io WAMP-over-WebSocket client protocol.
    """


class WampWebSocketClientFactory(websocket.WampWebSocketClientFactory):
    """
    Crossbar.io WAMP-over-WebSocket client factory.
    """

    protocol = WampWebSocketClientProtocol

    def buildProtocol(self, addr):
        self._proto = websocket.WampWebSocketClientFactory.buildProtocol(self, addr)
        return self._proto


class WampRawSocketClientProtocol(rawsocket.WampRawSocketClientProtocol):
    """
    Crossbar.io WAMP-over-RawSocket client protocol.
    """


class WampRawSocketClientFactory(rawsocket.WampRawSocketClientFactory):
    """
    Crossbar.io WAMP-over-RawSocket client factory.
    """

    protocol = WampRawSocketClientProtocol

    def __init__(self, factory, config):

        # transport configuration
        self._config = config

        # WAMP serializer
        #
        serid = config.get('serializer', 'json')

        if serid == 'json':
            # try JSON WAMP serializer
            try:
                from autobahn.wamp.serializer import JsonSerializer
                serializer = JsonSerializer()
            except ImportError:
                raise Exception("could not load WAMP-JSON serializer")

        elif serid == 'msgpack':
            # try MessagePack WAMP serializer
            try:
                from autobahn.wamp.serializer import MsgPackSerializer
                serializer = MsgPackSerializer()
                serializer._serializer.ENABLE_V5 = False  # FIXME
            except ImportError:
                raise Exception("could not load WAMP-MessagePack serializer")

        elif serid == 'cbor':
            # try CBOR WAMP serializer
            try:
                from autobahn.wamp.serializer import CBORSerializer
                serializer = CBORSerializer()
            except ImportError:
                raise Exception("could not load WAMP-CBOR serializer")

        else:
            raise Exception("invalid WAMP serializer '{}'".format(serid))

        rawsocket.WampRawSocketClientFactory.__init__(self, factory, serializer)


class WebSocketReverseProxyClientProtocol(websocket.WebSocketClientProtocol):

    log = make_logger()

    def onConnect(self, response):
        self.log.debug('WebSocketReverseProxyClientProtocol.onConnect(response={response})', response=response)
        # response={"peer": "tcp4:127.0.0.1:9000", "headers": {"server": "AutobahnPython/17.10.1", "upgrade": "WebSocket", "connection": "Upgrade", "sec-websocket-accept": "tJFVMTSzGypbxQb8GW1dK/QLMZQ="}, "version": 18, "protocol": null, "extensions": []}
        try:
            # accept = ConnectionAccept(subprotocol=response.protocol, headers=None)
            headers = {}
            accept = response.protocol, headers
            self.backend_on_connect.callback(accept)
        except:
            self.log.failure()

    def onOpen(self):
        self.log.debug('WebSocketReverseProxyClientProtocol.onOpen()')
        self.factory.frontend_protocol.onOpen()

    def onMessage(self, payload, isBinary):
        self.log.debug('WebSocketReverseProxyClientProtocol.onMessage(payload={payload}, isBinary={isBinary})',
                       payload='{}..'.format(
                           (binascii.b2a_hex(payload).decode() if isBinary else payload.decode('utf8'))[:16]),
                       isBinary=isBinary)
        self.factory.frontend_protocol.sendMessage(payload, isBinary)

    def onClose(self, wasClean, code, reason):
        self.log.debug(
            'WebSocketReverseProxyClientProtocol.onClose(wasClean={wasClean}, code={code}, reason={reason})',
            wasClean=wasClean,
            code=code,
            reason=reason)
        code = 1000
        self.factory.frontend_protocol.sendClose(code, reason)


class WebSocketReverseProxyClientFactory(websocket.WebSocketClientFactory):

    log = make_logger()

    protocol = WebSocketReverseProxyClientProtocol

    def __init__(self, *args, **kwargs):
        self.frontend_protocol = kwargs.pop('frontend_protocol', None)
        assert (self.frontend_protocol is not None)

        frontend_request = kwargs.pop('frontend_request', None)
        if frontend_request:
            protocols = frontend_request.protocols
            # headers = frontend_request.headers
        else:
            protocols = None
            # headers = None

        websocket.WebSocketClientFactory.__init__(self, *args, protocols=protocols, **kwargs)

        # set WebSocket options
        self.backend_config = self.frontend_protocol.backend_config
        set_websocket_options(self, self.backend_config.get('options', {}))


class WebSocketReverseProxyServerProtocol(websocket.WebSocketServerProtocol):
    """
    Server protocol to accept incoming (frontend) WebSocket connections and
    forward traffic to a backend WebSocket server.

    This protocol supports any WebSocket based subprotocol with text or
    binary payload.

    The WebSocket connection to the backend WebSocket server is configurable
    on the factory for this protocol.
    """

    log = make_logger()

    def onConnect(self, request):
        """
        Incoming (frontend) WebSocket connection accepted. Forward connect to
        backend WebSocket server.
        """
        self.log.debug('WebSocketReverseProxyServerProtocol.onConnect(request={request})', request=request)

        self.backend_config = self.factory.path_config['backend']

        self.backend_factory = WebSocketReverseProxyClientFactory(frontend_protocol=self,
                                                                  frontend_request=request,
                                                                  url=self.backend_config.get('url', None))
        self.backend_factory.noisy = False

        self.backend_protocol = None

        # create and connect client endpoint
        #
        endpoint = create_connecting_endpoint_from_config(self.backend_config['endpoint'], None, self.factory.reactor,
                                                          self.log)

        backend_on_connect = internet.defer.Deferred()

        # now, actually connect the client
        #
        d = endpoint.connect(self.backend_factory)

        def on_connect_success(proto):
            self.log.debug('WebSocketReverseProxyServerProtocol.onConnect(..): connected')
            proto.backend_on_connect = backend_on_connect
            self.backend_protocol = proto

        def on_connect_error(err):
            deny = ConnectionDeny(ConnectionDeny.SERVICE_UNAVAILABLE, 'WebSocket reverse proxy backend not reachable')
            backend_on_connect.errback(deny)

        d.addCallbacks(on_connect_success, on_connect_error)

        return backend_on_connect

    def onOpen(self):
        self.log.debug('WebSocketReverseProxyServerProtocol.onOpen()')

    def onMessage(self, payload, isBinary):
        if self.backend_protocol:
            self.log.debug(
                'WebSocketReverseProxyServerProtocol: forwarding WebSocket message from frontend connection to backend connection'
            )
            self.backend_protocol.sendMessage(payload, isBinary)
        else:
            self.log.warn(
                'WebSocketReverseProxyServerProtocol: received WebSocket message on frontend connection while there is no backend connection! dropping WebSocket message'
            )

    def onClose(self, wasClean, code, reason):
        if self.backend_protocol:
            self.log.debug(
                'WebSocketReverseProxyServerProtocol: forwarding close from frontend connection to backend connection')
            code = 1000
            self.backend_protocol.sendClose(code, reason)
        else:
            self.log.warn(
                'WebSocketReverseProxyServerProtocol: received WebSocket close on frontend connection while there is no backend connection! dropping WebSocket close'
            )


class WebSocketReverseProxyServerFactory(websocket.WebSocketServerFactory):
    """
    Reverse WebSocket proxy factory.

    This factory produces protocols that accept incoming WebSocket connections,
    connect to a backend WebSocket server and forward WebSocket messages received
    from both side to the other side.

    Reverse WebSocket proxy factories are then given to WebSocketResource instances
    in a Web transport resource tree.
    """

    protocol = WebSocketReverseProxyServerProtocol
    log = make_logger()

    def __init__(self, reactor, path_config):
        """

        :param path_config: The path configuration of the Web transport resource.
        :type path_config: dict
        """
        self.reactor = reactor
        self.path_config = path_config

        url = path_config.get('url', None)

        options = path_config.get('options', {})

        showServerVersion = options.get('show_server_version', True)
        if showServerVersion:
            server = "Crossbar/{}".format(crossbar.__version__)
        else:
            server = "Crossbar"

        externalPort = options.get('external_port', None)

        protocols = None
        headers = None

        websocket.WebSocketServerFactory.__init__(self,
                                                  reactor=reactor,
                                                  url=url,
                                                  protocols=protocols,
                                                  server=server,
                                                  headers=headers,
                                                  externalPort=externalPort)
        # set WebSocket options
        set_websocket_options(self, options)



###############################################################################
#
# Crossbar.io Shell
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import os
import re
import sys
import platform
import importlib
import hashlib
import json

import six
import click

# Notes:
#
# click does NOT allow to document CLI arguments! this is ridiculous, but true:
# https://github.com/pallets/click/issues/587
# to document, refer to arguments in UPPERCASE in the main command help. yeah, awesome.
# the justification is: "this has been the Unix way" - yeah, fuck that.

# this is for pyinstaller! otherwise it fails to find this dep.
# see: http://cffi.readthedocs.io/en/latest/cdef.html
import _cffi_backend  # noqa

# import and select network framework in txaio _before_ any further crossbar.shell imports
import txaio

txaio.use_twisted()  # noqa

from cfxdb.exporter import Exporter

from crossbar.shell import __version__, __build__  # noqa: E402
from crossbar.shell import command
from crossbar.shell.app import Application

_HEXKEY = re.compile(r'^[0-9a-fA-F]{64,64}$')

USAGE = """
Examples:
To start the interactive shell, use the "shell" command:

    cbf shell

You can run the shell under different user profile
using the "--profile" option:

    cbf --profile mister-test1 shell
"""

# the global, singleton app object
_app = Application()


def hl(text):
    if not isinstance(text, six.text_type):
        text = '{}'.format(text)
    return click.style(text, fg='yellow', bold=True)


class Config(object):
    """
    Command configuration object where we collect all the parameters,
    options etc for later processing.
    """
    def __init__(self, app, dotdir, profile, realm, role):
        self.app = app
        self.dotdir = dotdir
        self.profile = profile
        self.realm = realm
        self.role = role
        self.resource_type = None
        self.resource = None

    def __str__(self):
        return 'Config(resource_type={}, resource={})'.format(self.resource_type, self.resource)


@click.group(help="Crossbar.io Command Line", invoke_without_command=True)
@click.option(
    '--dotdir',
    envvar='CBF_DOTDIR',
    default=None,
    help="Set the dot directory (with config and profile) to be used",
)
@click.option(
    '--profile',
    envvar='CBF_PROFILE',
    default='default',
    help="Set the profile to be used",
)
@click.option(
    '--realm',
    envvar='CBF_REALM',
    default=None,
    help="Set the realm to join",
)
@click.option(
    '--role',
    envvar='CBF_ROLE',
    default=None,
    help="Set the role requested to authenticate as",
)
@click.option('--debug', is_flag=True, help='Enable debug output')
@click.pass_context
def cli(ctx, dotdir, profile, realm, role, debug):
    if debug:
        txaio.start_logging(level='info')

    ctx.obj = Config(_app, dotdir, profile, realm, role)

    # Allowing a command group to specifiy a default subcommand can be done using
    # https://github.com/click-contrib/click-default-group
    #
    # However, this breaks the click-repl integration for prompt-toolkit:
    #
    # https://github.com/pallets/click/issues/430#issuecomment-282015177
    #
    # Hence, we are using a different (probably less clean) trick - this works.
    #
    if ctx.invoked_subcommand is None:
        ctx.invoke(cmd_shell)


@cli.command(name='monitor', help='monitor master node')
@click.pass_context
def cmd_monitor(ctx):
    cfg = ctx.obj

    # key: userkey.UserKey
    # profile: config.Profile
    key, profile = ctx.obj.app.load_profile(profile=cfg.profile)

    from .monitor import run
    run(management_url=profile.url, privkey_file=profile.privkey)


@cli.command(name='version', help='print version information')
@click.pass_context
def cmd_version(ctx):
    def get_version(name_or_module):
        if isinstance(name_or_module, str):
            name_or_module = importlib.import_module(name_or_module)
        try:
            return name_or_module.__version__
        except AttributeError:
            return ''

    # Python (language)
    py_ver = '.'.join([str(x) for x in list(sys.version_info[:3])])

    # Python (implementation)
    if hasattr(sys, 'pypy_version_info'):
        pypy_version_info = getattr(sys, 'pypy_version_info')
        py_impl_str = '.'.join(str(x) for x in pypy_version_info[:3])
        py_ver_detail = "{}-{}".format(platform.python_implementation(), py_impl_str)
    else:
        py_ver_detail = platform.python_implementation()

    # Autobahn
    ab_ver = get_version('autobahn')

    # Pyinstaller (frozen EXE)
    py_is_frozen = getattr(sys, 'frozen', False)
    if py_is_frozen:
        m = hashlib.sha256()
        with open(sys.executable, 'rb') as fd:
            m.update(fd.read())
        fingerprint = m.hexdigest()
    else:
        fingerprint = None

    # Docker Compose
    try:
        import compose
    except ImportError:
        compose_ver = 'not installed'
    else:
        compose_ver = compose.__version__

    # Sphinx
    try:
        import sphinx
    except ImportError:
        sphinx_ver = 'not installed'
    else:
        sphinx_ver = sphinx.__version__

    platform_str = platform.platform(terse=True, aliased=True)

    click.echo()
    click.echo(hl("  Crossbar.io Shell") + ' - Command line tool for Crossbar.io')
    click.echo()
    click.echo('  {:<24}: {}'.format('Version', hl('{} (build {})'.format(__version__, __build__))))
    click.echo('  {:<24}: {}'.format('Platform', hl(platform_str)))
    click.echo('  {:<24}: {}'.format('Python (language)', hl(py_ver)))
    click.echo('  {:<24}: {}'.format('Python (implementation)', hl(py_ver_detail)))
    click.echo('  {:<24}: {}'.format('Autobahn', hl(ab_ver)))
    click.echo('  {:<24}: {}'.format('Docker Compose', hl(compose_ver)))
    click.echo('  {:<24}: {}'.format('Sphinx', hl(sphinx_ver)))
    click.echo('  {:<24}: {}'.format('Frozen executable', hl('yes' if py_is_frozen else 'no')))
    if py_is_frozen:
        click.echo('  {:<24}: {}'.format('Executable SHA256', hl(fingerprint)))
    click.echo()


@cli.command(name='init', help='create a new user profile / key-pair if none exists')
@click.option(
    '--yes',
    is_flag=True,
    default=False,
    help="Answer yes / use default for anything prompted",
)
@click.pass_context
def cmd_init(ctx, yes):
    cfg = ctx.obj
    cfg.yes_to_all = yes
    ctx.obj.app.run_context(ctx)


@cli.command(name='auth', help='authenticate user with Crossbar.io')
@click.option(
    '--code',
    default=None,
    help="Supply authentication code (received by email)",
)
@click.option(
    '--new-code',
    is_flag=True,
    default=False,
    help=  # noqa: E251
    "Request sending of a new authentication code (even though an old one is still pending)",
)
@click.option(
    '--yes',
    is_flag=True,
    default=False,
    help=  # noqa: E251
    "Answer yes / use default for anything prompted",
)
@click.pass_context
def cmd_auth(ctx, code, new_code, yes):
    cfg = ctx.obj
    cfg.code = code
    cfg.new_code = new_code
    cfg.yes_to_all = yes
    ctx.obj.app.run_context(ctx)


@cli.command(name='shell', help='run an interactive Crossbar.io Shell')
@click.pass_context
def cmd_shell(ctx):
    ctx.obj.app.run_context(ctx)


@cli.command(name='clear', help='clear screen')
def cmd_clear():
    click.clear()


@cli.command(name='help', help='general help')
@click.pass_context
def cmd_help(ctx):
    click.echo(ctx.parent.get_help())
    click.echo(USAGE)


@cli.group(name='set', help='change shell settings')
@click.pass_context
def cmd_set(ctx):
    pass


#
# set output-verbosity
#
@cmd_set.group(name='output-verbosity', help='command output verbosity')
@click.pass_context
def cmd_set_output_verbosity(ctx):
    pass


@cmd_set_output_verbosity.command(name='silent', help='swallow everything including result, but error messages')
@click.pass_context
def cmd_set_output_verbosity_silent(ctx):
    ctx.obj.app.set_output_verbosity('silent')


@cmd_set_output_verbosity.command(name='result-only', help='only output the plain command result')
@click.pass_context
def cmd_set_output_verbosity_result_only(ctx):
    ctx.obj.app.set_output_verbosity('result-only')


@cmd_set_output_verbosity.command(name='normal', help='output result and short run-time message')
@click.pass_context
def cmd_set_output_verbosity_normal(ctx):
    ctx.obj.app.set_output_verbosity('normal')


@cmd_set_output_verbosity.command(name='extended', help='output result and extended execution information.')
@click.pass_context
def cmd_set_output_verbosity_extended(ctx):
    ctx.obj.app.set_output_verbosity('extended')


#
# set output-format
#
@cmd_set.group(name='output-format', help='command output format')
@click.pass_context
def cmd_set_output_format(ctx):
    pass


def _make_set_output_format(output_format):
    @cmd_set_output_format.command(name=output_format, help='set {} output format'.format(output_format.upper()))
    @click.pass_context
    def f(ctx):
        ctx.obj.app.set_output_format(output_format)

    return f


for output_format in Application.OUTPUT_FORMAT:
    _make_set_output_format(output_format)


#
# set output-style
#
@cmd_set.group(name='output-style', help='command output style')
@click.pass_context
def cmd_set_output_style(ctx):
    pass


def _make_set_output_style(output_style):
    @cmd_set_output_style.command(name=output_style, help='set {} output style'.format(output_style.upper()))
    @click.pass_context
    def f(ctx):
        ctx.obj.app.set_output_style(output_style)

    return f


for output_style in Application.OUTPUT_STYLE:
    _make_set_output_style(output_style)


@cli.group(name='show', help='show resources')
@click.pass_context
def cmd_show(ctx):
    pass


@cmd_show.command(name='status', help='show domain or mrealm status')
@click.pass_context
def cmd_show_domain_status(ctx):
    _command = command.CmdGetDomainStatus(ctx.obj.realm)
    ctx.obj.app.run_context(ctx, _command)


@cmd_show.command(name='version', help='get domain controller software version')
@click.pass_context
def cmd_show_domain_version(ctx):
    _command = command.CmdGetDomainVersion(ctx.obj.realm)
    ctx.obj.app.run_context(ctx, _command)


@cmd_show.command(name='license', help='get domain software stack license')
@click.pass_context
def cmd_show_domain_license(ctx):
    _command = command.CmdGetDomainLicense(ctx.obj.realm)
    ctx.obj.app.run_context(ctx, _command)


@cmd_show.command(name='database', help='open and show embedded database details')
@click.argument('dbpath')
@click.option('--include-slots/--no-include-slots', default=False, type=bool, help='show database slots')
@click.pass_context
def cmd_show_database(ctx, dbpath, include_slots):
    exporter = Exporter(dbpath)
    exporter.print_config()
    exporter.print_stats(include_slots=include_slots)


@cli.group(name='export', help='export resources')
@click.pass_context
def cmd_export(ctx):
    pass


@cmd_export.command(name='database', help='export embedded database')
@click.argument('dbpath')
@click.option('--filename')
@click.option('--include-indexes/--no-include-indexes', default=False, type=bool, help='export index tables')
@click.option('--include-schemata',
              type=str,
              help='list of schemata to export (meta, globalschema, mrealmschema, xbr, xbrmm, xbrnetwork)')
@click.option('--exclude-tables', type=str)
@click.option('--use-json/--no-use-json', default=False, type=bool)
@click.option('--use-binary-hex-encoding/--no-use-binary-hex-encoding', default=False, type=bool)
@click.option('--quiet/--no-quiet', default=False, type=bool)
@click.pass_context
def cmd_export_database(ctx, dbpath, filename, include_indexes, include_schemata, exclude_tables, use_json,
                        use_binary_hex_encoding, quiet):
    if include_schemata:
        include_schemata = include_schemata.split(',')
    if exclude_tables:
        exclude_tables = exclude_tables.split(',')

    exporter = Exporter(dbpath)
    exporter.export_database(filename,
                             include_indexes=include_indexes,
                             include_schemata=include_schemata,
                             exclude_tables=exclude_tables,
                             use_json=use_json,
                             use_binary_hex_encoding=use_binary_hex_encoding,
                             quiet=quiet)


@cli.group(name='import', help='import resources')
@click.pass_context
def cmd_import(ctx):
    pass


@cmd_import.command(name='database', help='import embedded database')
@click.argument('dbpath')
@click.argument('filename')
@click.option('--include-indexes/--no-include-indexes', default=False, type=bool, help='import index tables')
@click.option('--include-schemata', type=str)
@click.option('--exclude-tables', type=str)
@click.option('--use-json/--no-use-json', default=False, type=bool)
@click.option('--use-binary-hex-encoding/--no-use-binary-hex-encoding', default=False, type=bool)
@click.option('--quiet/--no-quiet', default=False, type=bool)
@click.pass_context
def cmd_import_database(ctx, dbpath, filename, include_indexes, include_schemata, exclude_tables, use_json,
                        use_binary_hex_encoding, quiet):
    if include_schemata:
        include_schemata = include_schemata.split(',')
    if exclude_tables:
        exclude_tables = exclude_tables.split(',')

    exporter = Exporter(dbpath)
    exporter.print_stats()
    exporter.import_database(filename,
                             include_indexes=include_indexes,
                             include_schemata=include_schemata,
                             exclude_tables=exclude_tables,
                             use_json=use_json,
                             use_binary_hex_encoding=use_binary_hex_encoding,
                             quiet=quiet)
    exporter.print_stats()


@cli.group(name='add', help='add resources')
@click.pass_context
def cmd_add(ctx):
    pass


@cmd_add.command(name='principal', help='add a principal to an application realm')
@click.argument('realm')
@click.argument('principal')
@click.option('--config', type=str)
@click.pass_context
def cmd_add_principal(ctx, realm, principal, config=None):
    if config:
        config = json.loads(config)
    else:
        config = {}
    config['authid'] = principal
    cmd = command.CmdAddPrincipal(realm, principal, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_add.command(name='principal-credential', help='add a credential to a principal')
@click.argument('realm')
@click.argument('principal')
@click.option('--config', type=str)
@click.pass_context
def cmd_add_principal_credential(ctx, realm, principal, config=None):
    if config:
        config = json.loads(config)
    else:
        config = {}
    cmd = command.CmdAddPrincipalCredential(realm, principal, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_add.command(name='role-permission', help='add a permission to a role')
@click.argument('role')
@click.argument('uri')
@click.option('--config', type=str)
@click.pass_context
def cmd_add_role_permission(ctx, role, uri, config=None):
    if config:
        config = json.loads(config)
    else:
        config = {}
    config['uri'] = uri
    cmd = command.CmdAddRolePermission(role, uri, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_add.command(name='arealm-role', help='add a role to an application realm')
@click.argument('realm')
@click.argument('role')
@click.option('--config', type=str)
@click.pass_context
def cmd_add_realm_role(ctx, realm, role, config=None):
    if config:
        config = json.loads(config)
    cmd = command.CmdAddApplicationRealmRole(realm, role, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_add.command(name='routercluster-node', help='add a node to a routercluster')
@click.argument('cluster')
@click.argument('node', default='all')
@click.option('--config', type=str)
@click.pass_context
def cmd_add_routercluster_node(ctx, cluster, node, config=None):
    if config:
        config = json.loads(config)
    cmd = command.CmdAddRouterClusterNode(cluster, node, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_add.command(name='routercluster-workergroup', help='add a workergroup to a routercluster')
@click.argument('cluster')
@click.argument('workergroup')
@click.option('--config', type=str)
@click.pass_context
def cmd_add_routercluster_workergroup(ctx, cluster, workergroup, config=None):
    if config:
        config = json.loads(config)
    else:
        config = {}
    config['name'] = workergroup
    cmd = command.CmdAddRouterClusterWorkerGroup(cluster, workergroup, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_add.command(name='webcluster-node', help='add a node to a webcluster')
@click.argument('cluster')
@click.argument('node', default='all')
@click.option('--config', type=str)
@click.pass_context
def cmd_add_webcluster_node(ctx, cluster, node, config=None):
    if config:
        config = json.loads(config)
    cmd = command.CmdAddWebClusterNode(cluster, node, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_add.command(name='webcluster-service', help='add a web service to a webcluster')
@click.argument('cluster')
@click.argument('path')
@click.option('--config', type=str)
@click.pass_context
def cmd_add_webcluster_service(ctx, cluster, path, config=None):
    if config:
        webservice = json.loads(config)
    else:
        webservice = {}
        webservice['path'] = path
    cmd = command.CmdAddWebClusterService(cluster, path, webservice)
    ctx.obj.app.run_context(ctx, cmd)


@cli.group(name='create', help='create resources')
@click.pass_context
def cmd_create(ctx):
    pass


@cmd_create.command(name='mrealm', help='create a new management realm')
@click.argument('realm')
@click.pass_context
def cmd_create_management_realm(ctx, realm):
    cmd = command.CmdCreateManagementRealm(realm=realm)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_create.command(name='arealm', help='create a new application realm')
@click.argument('realm')
@click.option('--config', type=str)
@click.pass_context
def cmd_create_application_realm(ctx, realm, config=None):
    if config:
        config = json.loads(config)
    else:
        config = {}
    config['name'] = realm
    cmd = command.CmdCreateApplicationRealm(config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_create.command(name='role', help='create a new role')
@click.argument('role')
@click.option('--config', type=str)
@click.pass_context
def cmd_create_role(ctx, role, config=None):
    if config:
        config = json.loads(config)
    else:
        config = {}
    config['name'] = role
    cmd = command.CmdCreateRole(config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_create.command(name='routercluster', help='create a new routercluster')
@click.argument('cluster')
@click.option('--config', type=str)
@click.pass_context
def cmd_create_routercluster(ctx, cluster, config=None):
    if config:
        config = json.loads(config)
    else:
        config = {}
    config['name'] = cluster
    cmd = command.CmdCreateRouterCluster(config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_create.command(name='webcluster', help='create a new webcluster')
@click.argument('cluster')
@click.option('--config', type=str)
@click.pass_context
def cmd_create_webcluster(ctx, cluster, config=None):
    if config:
        config = json.loads(config)
    else:
        config = {}
    config['name'] = cluster
    cmd = command.CmdCreateWebCluster(config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_create.command(name='webservice', help='create a webservice within a webcluster')
@click.argument('cluster')
@click.argument('config')
@click.pass_context
def cmd_create_webservice(ctx, cluster, config):
    config = json.loads(config)
    cmd = command.CmdCreateWebService(cluster, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_create.command(name='docker-container', help='create a new Docker container on a node')
@click.argument('node')
@click.argument('image')
@click.option('--config', type=str)
@click.pass_context
def cmd_create_docker_container(ctx, node, image, config=None):
    if config:
        config = json.loads(config)
    else:
        config = {}
    cmd = command.CmdCreateDockerContainer(node, image, config)
    ctx.obj.app.run_context(ctx, cmd)


@cli.group(name='remove', help='remove resources')
@click.pass_context
def cmd_remove(ctx):
    pass


@cmd_remove.command(name='routercluster-node', help='remove a node from a routercluster')
@click.argument('cluster')
@click.argument('node')
@click.pass_context
def cmd_remove_routercluster_node(ctx, cluster, node):
    cmd = command.CmdRemoveRouterClusterNode(cluster, node)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_remove.command(name='routercluster-workergroup', help='remove a workergroup from a routercluster')
@click.argument('cluster')
@click.argument('workergroup')
@click.pass_context
def cmd_remove_routeercluster_workergroup(ctx, cluster, workergroup):
    cmd = command.CmdRemoveRouterClusterWorkerGroup(cluster, workergroup)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_remove.command(name='webcluster-node', help='remove a node from a webcluster')
@click.argument('cluster')
@click.argument('node')
@click.pass_context
def cmd_remove_webcluster_node(ctx, cluster, node):
    cmd = command.CmdRemoveWebClusterNode(cluster, node)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_remove.command(name='webcluster-service', help='remove a service from a webcluster')
@click.argument('cluster')
@click.argument('path')
@click.pass_context
def cmd_remove_webcluster_service(ctx, cluster, path):
    cmd = command.CmdRemoveWebClusterService(cluster, path)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_remove.command(name='arealm-principal', help='remove a principal from an application realm')
@click.argument('arealm')
@click.argument('principal')
@click.pass_context
def cmd_remove_arealm_principal(ctx, arealm, principal):
    cmd = command.CmdRemoveArealmPrincipal(arealm, principal)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_remove.command(name='principal-credential', help='remove credentials from a principal on an application realm')
@click.argument('arealm')
@click.argument('principal')
@click.argument('credential')
@click.pass_context
def cmd_remove_arealm_principal_credential(ctx, arealm, principal, credential):
    cmd = command.CmdRemoveArealmPrincipalCredential(arealm, principal, credential)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_remove.command(name='role-permission', help='remove a permission from a role')
@click.argument('role')
@click.argument('path')
@click.pass_context
def cmd_remove_role_permission(ctx, role, path):
    cmd = command.CmdRemoveRolePermission(role, path)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_remove.command(name='arealm-role', help='remove a role from an application realm')
@click.argument('arealm')
@click.argument('role')
@click.pass_context
def cmd_remove_arealm_role(ctx, arealm, role):
    cmd = command.CmdRemoveArealmRole(arealm, role)
    ctx.obj.app.run_context(ctx, cmd)


@cli.group(name='delete', help='delete resources')
@click.pass_context
def cmd_delete(ctx):
    pass


@cmd_delete.command(name='mrealm', help='Delete an existing management REALM by name.')
@click.argument('realm', type=str)
@click.option(
    '--cascade',
    is_flag=True,
    help=
    'Automatically unpair (but not delete) any nodes currently paired with and unassign (but not delete) any users currently assigned to the management realm to be deleted.'
)
@click.pass_context
def cmd_delete_management_realm(ctx, realm, cascade=False):
    cmd = command.CmdDeleteManagementRealm(realm, cascade)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_delete.command(name='routercluster', help='delete an existing webrouterclustercluster')
@click.argument('cluster')
@click.pass_context
def cmd_delete_routercluster(ctx, cluster):
    cmd = command.CmdDeleteRouterCluster(cluster)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_delete.command(name='webcluster', help='delete an existing webcluster')
@click.argument('cluster')
@click.pass_context
def cmd_delete_webcluster(ctx, cluster):
    cmd = command.CmdDeleteWebCluster(cluster)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_delete.command(name='arealm', help='delete an existing application realm')
@click.argument('arealm')
@click.option('--cascade', is_flag=True, help='Automatically delete dependent resources of the application realm.')
@click.pass_context
def cmd_delete_arealm(ctx, arealm, cascade=False):
    cmd = command.CmdDeleteApplicationRealm(arealm, cascade)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_delete.command(name='role', help='delete an existing application role')
@click.argument('role')
@click.pass_context
def cmd_delete_role(ctx, role):
    cmd = command.CmdDeleteRole(role)
    ctx.obj.app.run_context(ctx, cmd)


@cli.group(name='pair', help='pair nodes and devices')
@click.pass_context
def cmd_pair(ctx):
    pass


def _read_pubkey(pubkey):
    if not _HEXKEY.match(pubkey):
        fn = os.path.expanduser(pubkey)
        if not os.path.exists(fn):
            raise Exception('could not open node public key file {}'.format(fn))
        public_hex = None
        with open(fn, 'r') as f:
            data = f.read()
            for line in data.splitlines():
                if line.startswith('public-key-ed25519'):
                    public_hex = line.split(':')[1].strip()
                    break
        if not public_hex:
            raise Exception('no public key found in node public key file {}'.format(fn))
        pubkey = public_hex
    return pubkey


@cmd_pair.command(name='node', help='pair a node')
@click.argument('pubkey')
@click.argument('realm')
@click.argument('node_id')
@click.option('--authextra', type=str)
@click.pass_context
def cmd_pair_node(ctx, pubkey, realm, node_id, authextra=None):
    """

    :param ctx:
    :param pubkey: the public key of the node, a 32 bytes Ed25519 public key provided as a HEX string (64 characters),
        or alternatively a filename to read the public key from
    :param realm: management realm the node is to be paired with
    :param node_id: the ID of node assigned when the node is connecting
    :param authextra: authextra to be provided to the node when connecting, must be a JSON string
    :return:
    """
    assert type(pubkey) == str
    assert type(realm) == str
    assert type(node_id) == str
    assert authextra is None or type(authextra) == str

    pubkey = _read_pubkey(pubkey)

    if authextra:
        authextra = json.loads(authextra)

    cmd = command.CmdPairNode(pubkey=pubkey, realm=realm, node_id=node_id, authextra=authextra)
    ctx.obj.app.run_context(ctx, cmd)


@cli.group(name='unpair', help='unpair nodes and devices')
@click.pass_context
def cmd_unpair(ctx):
    pass


@cmd_unpair.command(name='node', help='unpair a node')
@click.argument('pubkey')
@click.pass_context
def cmd_unpair_node(ctx, pubkey):
    """

    :param ctx:
    :param pubkey:
    :return:
    """
    assert type(pubkey) == str

    pubkey = _read_pubkey(pubkey)

    cmd = command.CmdUnpairNode(pubkey=pubkey)
    ctx.obj.app.run_context(ctx, cmd)


@cli.group(name='start', help='start workers, components, ..')
@click.pass_context
def cmd_start(ctx):
    pass


# @cmd_start.command(name='worker', help='start a worker')
# @click.argument('node')
# @click.argument('worker')
# @click.argument('worker-type')
# @click.option('--options', help='worker options', default=None)
# @click.pass_context
# def cmd_start_worker(ctx, node, worker, worker_type, options=None):
#    cmd = command.CmdStartWorker(node, worker, worker_type, worker_options=options)
#    ctx.obj.app.run_context(ctx, cmd)
# from crossbar.shell.command import CmdStartContainerWorker, CmdStartContainerComponent


@cmd_start.command(name='docker-container', help='start a Docker container on a node')
@click.argument('node')
@click.argument('container')
@click.pass_context
def cmd_start_docker_container(ctx, node, container):
    cmd = command.CmdStartDockerContainer(node, container)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='container-worker', help='start a container worker')
@click.option('--process-title', help='worker process title (at OS level)', default=None)
@click.argument('node')
@click.argument('worker')
@click.pass_context
def cmd_start_container_worker(ctx, node, worker, process_title=None):
    cmd = command.CmdStartContainerWorker(node, worker, process_title=process_title)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='container-component', help='start a container component')
@click.argument('node')
@click.argument('worker')
@click.argument('component')
@click.pass_context
def cmd_start_container_component(ctx, node, worker, component):
    cmd = command.CmdStartContainerComponent(node, worker, component)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='router-worker', help='start a router worker')
@click.option('--process-title', help='worker process title (at OS level)', default=None)
@click.argument('node')
@click.argument('worker')
@click.pass_context
def cmd_start_router_worker(ctx, node, worker, process_title=None):
    cmd = command.CmdStartRouterWorker(node, worker, process_title=process_title)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='router-realm', help='start a router realm')
@click.argument('node')
@click.argument('worker')
@click.argument('realm')
@click.pass_context
def cmd_start_router_realm(ctx, node, worker, realm):
    cmd = command.CmdStartRouterRealm(node, worker, realm)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='router-transport', help='start a router transport')
@click.argument('node')
@click.argument('worker')
@click.argument('transport')
@click.option('--config', type=str)
@click.pass_context
def cmd_start_router_transport(ctx, node, worker, transport, config=None):
    if config:
        config = json.loads(config)
    cmd = command.CmdStartRouterTransport(node, worker, transport, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='webtransport-service', help='start a web-transport service')
@click.argument('node')
@click.argument('worker')
@click.argument('transport')
@click.argument('path')
@click.option('--config', type=str)
@click.pass_context
def cmd_start_web_transport_service(ctx, node, worker, transport, path, config=None):
    if config:
        config = json.loads(config)
    cmd = command.CmdStartWebTransportService(node, worker, transport, path, config)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='guest-worker', help='start a guest worker')
@click.argument('node')
@click.argument('worker')
@click.pass_context
def cmd_start_guest_worker(ctx, node, worker):
    cmd = command.CmdStartGuestWorker(node, worker)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='routercluster', help='start a routercluster')
@click.argument('cluster')
@click.pass_context
def cmd_start_routercluster(ctx, cluster):
    cmd = command.CmdStartRouterCluster(cluster)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='arealm', help='start an application realm on a router worker group')
@click.argument('arealm')
@click.argument('routercluster')
@click.argument('workergroup')
@click.argument('webcluster')
@click.pass_context
def cmd_start_arealm(ctx, arealm, routercluster, workergroup, webcluster):
    cmd = command.CmdStartApplicationRealm(arealm, routercluster, workergroup, webcluster)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_start.command(name='webcluster', help='start a webcluster')
@click.argument('cluster')
@click.pass_context
def cmd_start_webcluster(ctx, cluster):
    cmd = command.CmdStartWebCluster(cluster)
    ctx.obj.app.run_context(ctx, cmd)


@cli.group(name='stop', help='stop workers, components, ..')
@click.pass_context
def cmd_stop(ctx):
    pass


@cmd_stop.command(name='docker-container', help='stop a Docker container running on a node')
@click.argument('node')
@click.argument('container')
@click.pass_context
def cmd_stop_docker_container(ctx, node, container):
    cmd = command.CmdStopDockerContainer(node, container)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_stop.command(name='worker', help='stop a worker')
@click.argument('node')
@click.argument('worker')
@click.pass_context
def cmd_stop_worker(ctx, node, worker):
    cmd = command.CmdStopWorker(node, worker)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_stop.command(name='router-realm', help='stop a router realm')
@click.argument('node')
@click.argument('worker')
@click.argument('realm')
@click.pass_context
def cmd_stop_router_realm(ctx, node, worker, realm):
    cmd = command.CmdStopRouterRealm(node, worker, realm)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_stop.command(name='router-transport', help='stop a router transport')
@click.argument('node')
@click.argument('worker')
@click.argument('transport')
@click.pass_context
def cmd_stop_router_transport(ctx, node, worker, transport):
    cmd = command.CmdStopRouterTransport(node, worker, transport)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_stop.command(name='routercluster', help='stop a routercluster')
@click.argument('cluster')
@click.pass_context
def cmd_stop_routercluster(ctx, cluster):
    cmd = command.CmdStopRouterCluster(cluster)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_stop.command(name='webcluster', help='stop a webcluster')
@click.argument('cluster')
@click.pass_context
def cmd_stop_webcluster(ctx, cluster):
    cmd = command.CmdStopWebCluster(cluster)
    ctx.obj.app.run_context(ctx, cmd)


@cli.group(name='list', help='list resources')
@click.pass_context
def cmd_list(ctx):
    pass


@cmd_list.command(name='mrealms', help='list management realms')
@click.option('--names/--no-names', default=False, type=bool, help='return node names (authid) instead of object IDs')
@click.pass_context
def cmd_list_management_realms(ctx, names):
    cmd = command.CmdListManagementRealms(names=names)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='nodes', help='list nodes')
@click.option('--names/--no-names', default=False, type=bool, help='return node names (authid) instead of object IDs')
@click.option(
    '--online',
    is_flag=True,
    default=False,
    help=  # noqa: E251
    "List only nodes that are currently online",
)
@click.option(
    '--offline',
    is_flag=True,
    default=False,
    help=  # noqa: E251
    "List only nodes that are currently offline",
)
@click.pass_context
def cmd_list_nodes(ctx, names=None, online=None, offline=None):
    assert not (online and offline)
    cmd = command.CmdListNodes(online=online, offline=offline, names=names)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='workers', help='list workers')
@click.argument('node')
@click.pass_context
def cmd_list_workers(ctx, node):
    cmd = command.CmdListWorkers(node)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='router-realms', help='list router realms')
@click.argument('node')
@click.argument('worker')
@click.pass_context
def cmd_list_router_realms(ctx, node, worker):
    cmd = command.CmdListRouterRealms(node, worker)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='router-transports', help='list router transports')
@click.argument('node')
@click.argument('worker')
@click.pass_context
def cmd_list_router_transports(ctx, node, worker):
    cmd = command.CmdListRouterTransports(node, worker)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='arealms', help='list application realms')
@click.option('--names/--no-names',
              default=False,
              type=bool,
              help='return application realm names instead of object IDs')
@click.pass_context
def cmd_list_arealms(ctx, names):
    cmd = command.CmdListARealms(names=names)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='arealm-roles', help='list roles associated with application realm')
@click.argument('arealm')
@click.option('--names/--no-names', default=False, type=bool, help='return role names instead of object IDs')
@click.pass_context
def cmd_list_arealm_roles(ctx, arealm, names):
    cmd = command.CmdListARealmRoles(arealm, names=names)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='roles', help='list roles')
@click.option('--names/--no-names', default=False, type=bool, help='return role names instead of object IDs')
@click.pass_context
def cmd_list_roles(ctx, names):
    cmd = command.CmdListRoles(names=names)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='role-permissions', help='list role permissions')
@click.argument('role')
@click.pass_context
def cmd_list_role_permissions(ctx, role):
    cmd = command.CmdListRolePermissions(role)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='principals', help='list principals')
@click.argument('arealm')
@click.option('--names/--no-names', default=False, type=bool, help='return principals names instead of object IDs')
@click.pass_context
def cmd_list_principals(ctx, arealm, names):
    cmd = command.CmdListPrincipals(arealm, names=names)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='principal-credentials', help='list credentials of a principal')
@click.argument('arealm')
@click.argument('principal')
@click.pass_context
def cmd_list_principal_credentials(ctx, arealm, principal):
    cmd = command.CmdListPrincipalCredentials(arealm, principal)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='routerclusters', help='list routerclusters')
@click.option('--names/--no-names', default=False, type=bool, help='return routerclusters names instead of object IDs')
@click.pass_context
def cmd_list_routerclusters(ctx, names):
    cmd = command.CmdListRouterClusters(names=names)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='routercluster-nodes', help='list nodes of a routercluster')
@click.argument('cluster')
@click.option('--names/--no-names',
              default=False,
              type=bool,
              help='return routercluster nodes names (= authid) instead of object IDs')
@click.option('--filter-status',
              default=None,
              type=str,
              help='filter nodes returned by node status given, eg "online"')
@click.pass_context
def cmd_list_routercluster_nodes(ctx, cluster, names, filter_status):
    cmd = command.CmdListRouterClusterNodes(cluster, names=names, filter_status=filter_status)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='routercluster-workergroups', help='list workergroups of a routercluster')
@click.argument('cluster')
@click.option('--names/--no-names',
              default=False,
              type=bool,
              help='return routercluster workergroup names instead of object IDs')
@click.option('--filter-status',
              default=None,
              type=str,
              help='filter workergroups returned by workergroup status given, eg "online"')
@click.pass_context
def cmd_list_routercluster_workergroups(ctx, cluster, names, filter_status):
    cmd = command.CmdListRouterClusterWorkerGroups(cluster, names=names, filter_status=filter_status)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='webclusters', help='list webclusters')
@click.option('--names/--no-names', default=False, type=bool, help='return webcluster names instead of object IDs')
@click.pass_context
def cmd_list_webclusters(ctx, names):
    cmd = command.CmdListWebClusters(names=names)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='webcluster-nodes', help='list nodes of a webcluster')
@click.argument('cluster')
@click.option('--names/--no-names',
              default=False,
              type=bool,
              help='return webcluster nodes names (= authid) instead of object IDs')
@click.option('--filter-status',
              default=None,
              type=str,
              help='filter nodes returned by node status given, eg "online"')
@click.pass_context
def cmd_list_webcluster_nodes(ctx, cluster, names, filter_status):
    cmd = command.CmdListWebClusterNodes(cluster, names=names, filter_status=filter_status)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='webcluster-services', help='list webservices of a webcluster')
@click.argument('cluster')
@click.pass_context
def cmd_list_webcluster_webservices(ctx, cluster):
    cmd = command.CmdListWebClusterWebService(cluster)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='docker-images', help='list Docker images available on a node')
@click.argument('node')
@click.pass_context
def cmd_list_docker_images(ctx, node):
    cmd = command.CmdListDockerImages(node)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_list.command(name='docker-containers', help='list Docker containers on a node')
@click.argument('node')
@click.pass_context
def cmd_list_docker_containers(ctx, node):
    cmd = command.CmdListDockerContainers(node)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='fabric', help='show fabric')
@click.pass_context
def cmd_show_fabric(ctx):
    cmd = command.CmdShowFabric()
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='mrealm', help='show management realm (for domains)')
@click.argument('realm', default='any')
@click.pass_context
def cmd_show_mrealm(ctx, realm):
    cmd = command.CmdShowManagementRealm(realm)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='docker', help='show Docker information')
@click.argument('node')
@click.pass_context
def cmd_show_docker(ctx, node):
    cmd = command.CmdShowDocker(node)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='docker-image', help='show Docker image information')
@click.argument('node')
@click.argument('image')
@click.pass_context
def cmd_show_docker_image(ctx, node, image):
    cmd = command.CmdShowDockerImage(node, image)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='docker-container', help='show Docker container information')
@click.argument('node')
@click.argument('container')
@click.pass_context
def cmd_show_docker_container(ctx, node, container):
    cmd = command.CmdShowDockerContainer(node, container)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='node', help='show node')
@click.argument('node', default='all')
@click.pass_context
def cmd_show_node(ctx, node):
    cmd = command.CmdShowNode(node)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='worker', help='show worker')
@click.argument('node')
@click.argument('worker')
@click.pass_context
def cmd_show_worker(ctx, node, worker):
    cmd = command.CmdShowWorker(node, worker)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='transport', help='show transport (for router workers)')
@click.argument('node')
@click.argument('worker')
@click.argument('transport')
@click.pass_context
def cmd_show_transport(ctx, node, worker, transport):
    cmd = command.CmdShowTransport(node, worker, transport)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='realm', help='show realm (for router workers)')
@click.argument('node')
@click.argument('worker')
@click.argument('realm')
@click.pass_context
def cmd_show_realm(ctx, node, worker, realm):
    cmd = command.CmdShowRealm(node, worker, realm)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='component', help='show component (for container and router workers)')
@click.argument('node')
@click.argument('worker')
@click.argument('component')
@click.pass_context
def cmd_show_component(ctx, node, worker, component):
    cmd = command.CmdShowComponent(node, worker, component)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='arealm', help='show application realm')
@click.argument('arealm', default='all')
@click.pass_context
def cmd_show_arealm(ctx, arealm):
    cmd = command.CmdShowApplicationRealm(arealm)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='principal', help='show principal')
@click.argument('realm')
@click.argument('principal')
@click.pass_context
def cmd_show_principal(ctx, realm, principal):
    cmd = command.CmdShowPrincipal(realm, principal)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='role', help='show role')
@click.argument('role')
@click.pass_context
def cmd_show_role(ctx, role):
    cmd = command.CmdShowRole(role)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='role-permission', help='show role permission')
@click.argument('role')
@click.argument('uri')
@click.pass_context
def cmd_show_role_permission(ctx, role, uri):
    cmd = command.CmdShowRolePermission(role, uri)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='arealm-role', help='show arealm-role association')
@click.argument('arealm')
@click.argument('role')
@click.pass_context
def cmd_show_arealm_role(ctx, arealm, role):
    cmd = command.CmdShowARealmRole(arealm, role)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='routercluster', help='show routercluster')
@click.argument('cluster', default='all')
@click.pass_context
def cmd_show_routercluster(ctx, cluster):
    cmd = command.CmdShowRouterCluster(cluster)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='routercluster-node', help='show routercluster node')
@click.argument('cluster')
@click.argument('node', default='all')
@click.pass_context
def cmd_show_routercluster_node(ctx, cluster, node):
    cmd = command.CmdShowRouterClusterNode(cluster, node)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='routercluster-workergroup', help='show routercluster workergroup')
@click.argument('cluster')
@click.argument('workergroup')
@click.pass_context
def cmd_show_routercluster_workergroup(ctx, cluster, workergroup):
    cmd = command.CmdShowRouterClusterWorkerGroup(cluster, workergroup)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='webcluster', help='show webcluster')
@click.argument('cluster', default='all')
@click.pass_context
def cmd_show_webcluster(ctx, cluster):
    cmd = command.CmdShowWebCluster(cluster)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='webcluster-node', help='show webcluster node')
@click.argument('cluster')
@click.argument('node', default='all')
@click.pass_context
def cmd_show_webcluster_node(ctx, cluster, node):
    cmd = command.CmdShowWebClusterNode(cluster, node)
    ctx.obj.app.run_context(ctx, cmd)
    ctx.obj.app.run_context(ctx, cmd)


@cmd_show.command(name='webcluster-service', help='show webcluster webservice')
@click.argument('cluster')
@click.argument('webservice', default=None)
@click.pass_context
def cmd_show_webcluster_service(ctx, cluster, webservice):
    cmd = command.CmdShowWebClusterWebService(cluster, webservice)
    ctx.obj.app.run_context(ctx, cmd)


@cli.command(name='current', help='currently selected resource')
@click.pass_context
def cmd_current(ctx):
    _app.print_selected()


@cli.group(name='select', help='change current resource')
@click.pass_context
def cmd_select(ctx):
    pass


@cmd_select.command(name='node', help='change current node')
@click.argument('resource')
@click.pass_context
def cmd_select_node(ctx, resource):
    _app.current_resource_type = 'node'
    _app.current_resource = resource
    _app.print_selected()


@cmd_select.command(name='worker', help='change current worker')
@click.argument('resource')
@click.pass_context
def cmd_select_worker(ctx, resource):
    _app.current_resource_type = 'worker'
    _app.current_resource = resource
    _app.print_selected()


@cmd_select.command(name='transport', help='change current transport')
@click.argument('resource')
@click.pass_context
def cmd_select_transport(ctx, resource):
    _app.current_resource_type = 'transport'
    _app.current_resource = resource
    _app.print_selected()


def run():
    """
    Main entry point into CLI.
    """
    cli()  # pylint: disable=E1120


if __name__ == '__main__':
    run()
